
/*
FILE:	const.h
*/

# define	INFINITY	-1	/* infinity */

/* hard limits */

# define	MAX_PRIO	10000	/* maximum priority for a transition */
# define	MAX_SYMLEN	20	/* maximum symbol name length */
# define	MAX_TOKEN	126	/* max number of tokens in a place */
# define	MAX_TOKENplus1	127	/* MAX_TOKEN + 1 */
# define	MAX_CHAR	256	/* max number of different chars */
# define	MAX_ARRAY	1024	/* max n1 or n2 for an array */
# define	MAX_FLOAT	10e+30	/* infinity ... */

# define	HSH_SIZE	1229	/* symbol table size */
# define	HSH_FULL	1001	/* too many symbols */
# define	HSH_JUMP	1327	/* used for the overflows */

/* values returned by some functions */

# define	RES_NOERR	0	/* no error encountered */
# define	RES_ERROR	-1	/* error encountered */
# define        RES_EOF         1       /* eof encountered */


/* values used when comparing two markings */

# define	CMP_LT		0	/* lower than */
# define	CMP_GT		1	/* greater than */
# define	CMP_EQ		2	/* equal */

/* values used to classify the MC-transitions */

# define	MCT_ABS		-1	/* return to M0 form absmark */
# define	MCT_PTH		-2	/* immediate path ending with timed */

/* values used to classify symbols */

# define	SYM_PLACE	'p'
# define	SYM_TRANS	't'
# define	SYM_PARAM	'r'
# define	SYM_NOVAL	'X'

/* values used to classify the markings */

# define	MAR_TAN		't'	/* tangible marking */
# define	MAR_ABS		'a'	/* absorbing marking */
# define	MAR_VAN		'v'	/* vanishing marking */
# define	MAR_VLO		'l'	/* vanishing (in loop) marking */
# define	MAR_NOVAL	'X'

/* values used to classify transition distributions */

# define	DIS_IMM		'i'	/* immediate, (time = 0) */
# define	DIS_EXP		'e'	/* exponential */
# define	DIS_NOVAL	'X'

/* values used to classify marking dependency of transition */

# define	DEP_NO		'n'	/* constant */
# define	DEP_PL		'p'	/* infinite server on one place */
# define	DEP_MK		'm'	/* general marking dependent */
# define	DEP_NOVAL	'X'

/* values used in the predefined cspl functions */

# define	ALL		-10000	/* != from any place/trans index */
# define	SUM		-10001	/* != from any place/trans index */
# define	MAX		-10002	/* != from any place/trans index */
# define	MIN		-10003	/* != from any place/trans index */

/* values used to classify arcs */

# define	ARC_I		'i'	/* input arc */
# define	ARC_O		'o'	/* output arc */
# define	ARC_H		'h'	/* inhibitor arc */
# define	ARC_NOVAL	'X'
# define	VARIABLE	MAX_TOKENplus1	/* != from any multiplicity */

/* values used to index the array "FOpt". */

# define	IOP_PR_MARK_ORDER	0
/* order when printing the reach set: VAL_CANONIC,VAL_LEXICAL,VAL_MATRIX */

# define	IOP_PR_MERG_MARK	1
/* merge vanishing and tangible markings in printing rset: VAL_YES,VAL_NO */

# define	IOP_PR_FULL_MARK	2
/* print the markings in full format: VAL_YES,VAL_NO */

# define	IOP_PR_RSET	 	3
/* print the reach set: VAL_YES,VAL_NO,VAL_TANGIBLE */

# define	IOP_PR_RGRAPH		4
/* print the reach graph: VAL_YES,VAL_NO */

# define	IOP_MC			5
/* type of the Markov Chain: VAL_DTMC,VAL_CTMC */

# define	IOP_PR_MC		6
/* print the Markov Chain: VAL_YES,VAL_NO */

# define	IOP_PR_MC_ORDER		7
/* order when printing the Markov Chain: VAL_FROMTO,VAL_TOFROM */

# define	IOP_OK_ABSMARK		8
/* absorbing markings are acceptable: VAL_YES,VAL_NO */

# define	IOP_OK_VANLOOP		9
/* loops of vanishing markings are acceptable: VAL_YES,VAL_NO */

# define	IOP_OK_TRANS_M0		10
/* a transient initial marking is acceptable: VAL_YES,VAL_NO */

# define	IOP_DEBUG		11
/* output useful info during rgraph construction: VAL_YES,VAL_NO */

# define	IOP_USENAME		12
/* use the name instead of the number when printing reach graph info */

# define	IOP_METHOD		13
/* the solution method: VAL_GASEI,VAL_SSSOR */

# define	IOP_ITERATIONS		14
/* maximum number of iterations for the solution method */

# define	IOP_PR_PROB		15
/* print the Markov chain state probabilities: VAL_YES,VAL_NO */

# define	IOP_SENSITIVITY		16
/* print the Markov chain state probabilities: VAL_YES,VAL_NO */

# define	N_IOP			17
/* number of integer options */


/* values used to index the array "FOpt". */

# define	FOP_ABS_RET_M0		N_IOP+0
/* rate from absorbing markings to initial marking, if positive */

# define	FOP_PRECISION		N_IOP+1
/* required precision for the numerical solution */

# define	N_FOP	2
/* number of floating point options */


/* predefined values for the ioptions */

# define	VAL_NO		0
# define	VAL_YES		1
# define	VAL_TANGIBLE	3
# define	VAL_CANONIC	4
# define	VAL_LEXICAL	5
# define	VAL_MATRIX	6
# define	VAL_TOFROM	7
# define	VAL_FROMTO	8
# define	VAL_GASEI	9
# define	VAL_SSSOR	11
# define	VAL_TSUNIF	12	
# define	VAL_DTMC	13
# define	VAL_CTMC	14
# define	VAL_NOVAL	15

# define	USE_NAME	VAL_YES
# define	USE_NUMBER	VAL_NO

/* underflow value used in uniformization */
# define        UNDFLOW 1.0e-30


