
/*LINTLIBRARY*/

/*************************************************************************

FILE:	cspl.c

It implements the PN language.

FUNCTIONS CONTAINED:
	hash_look()
	insert_symb()
	search_symb()
	iopt()
	fopt()
	Cplace()
	findplace()
	Cinit()
	Ctrans()
	findtrans()
	Cpriority()
	Cenabling()
	Cdistr()
	Carc()
	install()
	internal_enabled()
	internal_rate()
	Cmark()
	Cenabled()
	Crate()
	input()
	expected()
	cum_expected()
	accumulated()
	hold_cond()
	pr_std_average()
	pr_std_cum_average()
	pr_net_info()
	pr_rg_info()
	pr_mc_info()
	pr_expected()
	pr_cum_expected()
	pr_accumulated()
	pr_hold_cond()
	Cset_prob0()
*/

# include	<stdio.h>
# include	<stdlib.h>
# include	<ctype.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"
# include	"var.h"
# include	"spsor.h"
# include	"net.h"

extern	char	*pr_place();
extern	char	*pr_trans();

int	Retval = RES_NOERR;	/* set to error code if something goes wrong */

int	Phase = -1;		/* needed because "net" is called more than once */

ST	Symtable[HSH_SIZE];	/* hash table */

char	Charval[MAX_CHAR];	/* table lookup for char hash values */

int	Prob0flag = VAL_NO;	/* if YES, initial probabilities were set */

double	*Prob0 = NULL;		/* initial probabilities */

# define	CHECK_A(n,f,l,a)	\
{	switch (n) {\
	case ALL:\
		f = 0;\
		l = a - 1;\
		break;\
	case SUM:\
	case MIN:\
	case MAX:\
		fprintf(stderr,"ERROR: illegal SUM, MIN, or MAX usage\n");\
		Retval = RES_ERROR;\
		return;\
	default:\
		f = l = n;\
		break;\
	}\
}

# define	CHECK_SMM(n,f,l,a,flg)	\
{	switch (n) {\
	case ALL:\
		fprintf(stderr,"ERROR: illegal ALL usage\n");\
		Retval = RES_ERROR;\
		return(0);\
	case SUM:\
	case MIN:\
	case MAX:\
		f = 0;\
		l = a - 1;\
		flg = n;\
		break;\
	default:\
		f = l = n;\
		flg = SUM;\
		break;\
	}\
}\

# define	SWITCHFLG(f,l,i,v1,v2,ACTION,flag,max)	\
{	switch (flag) {\
	case SUM:\
		v1 = 0;\
		for (i = f; i <= l; i++) {\
			ACTION\
			v1 += v2;\
		}\
		break;\
	case MIN:\
		v1 = max;\
		for (i = f; i <= l; i++) {\
			ACTION\
			if (v1 > v2) \
				v1 = v2;\
		}\
		break;\
	case MAX:\
		v1 = 0;\
		for (i = f; i <= l; i++) {\
			ACTION\
			if (v1 < v2)\
				v1 = v2;\
		}\
		break;\
	}\
}

int	hash_look(symb)
char	*symb;		/* the symbol to be searched */

/*************************************************************************
It looks in the symbol table for "symb".
RETURNS:
the index in the symbol table of the already existing "symb", or of the place
where symb should go.
*/

{
static	int	Hashfull = 0;
	int	i,hashval;
	char	*p1,*p2;

	/* compute hash value */
	hashval = Charval[symb[0]];
	if (symb[1] != '\0') {
		hashval += (Charval[symb[1]] << 6);
		if (symb[2] != '\0') {
			hashval += (Charval[symb[2]] << 12);
			if (symb[3] != '\0') {
				hashval += (Charval[symb[3]] << 18);
				if (symb[4] != '\0') {
					hashval += (Charval[symb[4]] << 24);
					for (i = 5; symb[i] != '\0'; i++) {
						hashval += symb[i];
					}
				}
			}
		}
	}
	hashval %= HSH_SIZE;
	/* check if symb is already present */
	while (Symtable[hashval].type != SYM_NOVAL) {
		/* compare the two strings */
		for (p1 = Symtable[hashval].symb, p2 = symb; ; p1++, p2++)
			if (*p1 == '\0' || *p2 == '\0' || *p1 != *p2) 
				break;
		if (*p1 == *p2)
			return(hashval);
		hashval = (hashval + HSH_JUMP) % HSH_SIZE;
	}
	if (++Hashfull == HSH_FULL) {
		fprintf(stderr,"ERROR: hash table overflow, too many symbols\n");
		exit(ER_CODE);
	}
	return(hashval);
}

int	insert_symb(name,type,numb)
char	*name;		/* the name whose validity is to be checked */
int	type;		/* the "type" field to be inserted */
int	numb;		/* the "numb" field to be inserted */

/*************************************************************************
Check the legality of a name. If legal, check if already existing.
If not existing, insert it.
RETURNS:
RES_ERROR if illegal or already present, RES_NOERR otherwise.
*/

{
extern	char	*get_SY();
	int	i,pos;

	/* check legality */
	if (Charval[name[0]] == 0 || name[0] == '_')
		return(RES_ERROR);
	for (i = 1; name[i] != '\0'; i++) {
		if (Charval[name[i]] == 0 || i == MAX_SYMLEN)
			return(RES_ERROR);
	}
	/* check if name is already present */
	pos = hash_look(name);
	if (Symtable[pos].type != SYM_NOVAL)
		return(RES_ERROR);
	/* insert name */
	Symtable[pos].symb = get_SY((unsigned int)(i+1));
	for (i = 0; (Symtable[pos].symb[i] = name[i]) != '\0'; i++);
	Symtable[pos].type = type;
	Symtable[pos].numb = numb;
	return(RES_NOERR);
}

int	search_symb(name)
char	*name;		/* name to be matched */

/*************************************************************************
Searches "Symtable" for a match with symbol "name".
RETURNS:
RES_ERROR if no match is found, the index in "Symtable" if found.
*/

{
	int	pos;

	/* check if name is already present */
	pos = hash_look(name);
	return((Symtable[pos].type == SYM_NOVAL) ? RES_ERROR : pos);
}

void	iopt(opt,val)
int	opt;		/* the index of the option */
int	val;		/* the value of the option */

/*************************************************************************
It sets the value of integer option "opt" to "val".
The value "val" is an integer number.
*/

{
	int	i;

	if (Phase != 0) {
		fprintf(stderr,"ERROR: 'iopt' must be called in 'parameters'\n");
		Retval = RES_ERROR;
		return;
	}
	if (opt < 0 || opt >= N_IOP) {
		fprintf(stderr,"ERROR: 'iopt' illegal option\n");
		Retval = RES_ERROR;
		return;
	}
	if (opt == IOP_ITERATIONS) {	/* just check if negative */
		if (val >= 0) {
			IOpt[opt].val = val;
			return;
		}
	} else {			/* check if legal value */
		for (i = 0; IOpt[opt].leg[i] != VAL_NOVAL
			&& IOpt[opt].leg[i] != val; i++);
		if (IOpt[opt].leg[i] != VAL_NOVAL) {
			IOpt[opt].val = val;
			return;
		}
	}
	fprintf(stderr,"ERROR: illegal value for option %s\n",IOpt[opt].str);
	Retval = RES_ERROR;
	return;
} /* iopt() */

void	fopt(opt,val)
int	opt;		/* the index of the option */
double	val;		/* the value of the option */

/*************************************************************************
It sets the value of floating point option "opt" to "val".
The value "val" is a real number.
*/

{

	if (Phase != 0) {
		fprintf(stderr,"ERROR: 'fopt' must be called in 'parameters'\n");
		Retval = RES_ERROR;
		return;
	}
	if (opt < N_IOP || opt >= N_IOP + N_FOP) {
		fprintf(stderr,"ERROR: 'fopt' illegal option\n");
		Retval = RES_ERROR;
		return;
	}
	if (val < 0) {
		fprintf(stderr,"ERROR: negative value for %s\n",FOpt[opt].str);
		Retval = RES_ERROR;
		return;
	}
	FOpt[opt].val = val;
} /* fopt() */

void	Cplace(name,n1,n2)
char	name[];		/* the name of the place */
int	n1;		/* number on 1st dimension */
int	n2;		/* number on 2nd dimension */

/*************************************************************************
It defines an array of places with names "name.0.0".."name.n1-1.n2-1".
*/

{
	char	*aux;
	int	i1,i2;

	switch (Phase) {
	case 0:
		return;
	case 1:
		if (insert_symb(name,SYM_PLACE,Nplace) == RES_ERROR) {
			fprintf(stderr,"ERROR: illegal place name ('%s')\n",name);
			Retval = RES_ERROR;
			return;
		}
		if (n1 <= 0 || n2 <= 0 || n1 > MAX_ARRAY || n2 > MAX_ARRAY) {
			fprintf(stderr,"ERROR: invalid array size for place '%s' \n",name);
			Retval = RES_ERROR;
			return;
		}
		/* increase the count */
		Nplace += (n1 * n2);
		break;
	case 2:
		aux = Arrplace[Nplace].name;
		for (i1 = 0; i1 < n1; i1++)
			for (i2 = 0; i2 < n2; i2++) {
				Arrplace[Nplace].name = aux;
				Arrplace[Nplace].i1 = i1;
				Arrplace[Nplace].i2 = i2;
				Arrplace[Nplace].n1 = n1;
				Arrplace[Nplace].n2 = n2;
				Arrplace[Nplace].initial = 0;
				Arrplace[Nplace].num = Nplace;
				Nplace++;
			}
		break;
	case 3:
		break;
	}
}

int	findplace(name,n1,n2)
char	name[];		/* the name of the place */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
RETURNS:
the index of place "name.n1.n2" in Arrplace, or RES_ERROR.
*/

{
	int	numb,index;

	index = search_symb(name);
	if (index == RES_ERROR || Symtable[index].type != SYM_PLACE) {
		fprintf(stderr,"ERROR: place '%s' undefined\n",name);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	numb = Symtable[index].numb;
	if (n1 !=  ALL && n1 != SUM && n1 != MIN && n1 != MAX &&
	(n1 < 0 || n1 >= Arrplace[numb].n1)) {
		fprintf(stderr,"ERROR: illegal first index for place '%s' (%d)\n",name,n1);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	if (n2 !=  ALL && n2 != SUM && n2 != MIN && n2 != MAX &&
	(n2 < 0 || n2 >= Arrplace[numb].n2)) {
		fprintf(stderr,"ERROR: illegal second index for place '%s' (%d)\n",name,n2);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	return(numb);
}

int	findtrans(name,n1,n2)
char	name[];		/* the name of the transition */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
RETURNS:
the index of trans "name.n1.n2" in Arrtrans, or RES_ERROR.
*/

{
	int	index,numb;

	index = search_symb(name);
	if (index == RES_ERROR || Symtable[index].type != SYM_TRANS) {
		fprintf(stderr,"ERROR: transition '%s' undefined\n",name);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	numb = Symtable[index].numb;
	if (n1 !=  ALL && n1 != SUM && n1 != MIN && n1 != MAX &&
	(n1 < 0 || n1 >= Arrtrans[numb].n1)) {
		fprintf(stderr,"ERROR: illegal first index for transition '%s' (%d)\n",name,n1);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	if (n2 !=  ALL && n2 != SUM && n2 != MIN && n2 != MAX &&
	(n2 < 0 || n2 >= Arrtrans[numb].n2)) {
		fprintf(stderr,"ERROR: illegal second index for transition '%s' (%d)\n",name,n2);
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	return(numb);
}

void	Cinit(name,n1,n2,init)
char	name[];		/* the name of the place */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */
int	init;		/* initial number of teokens */

/*************************************************************************
Sets to "init" the initial number of tokens in place "name.n1.n2".
*/

{
	int	f1,l1,f2,l2,i1,i2,firstindex,index;
	char	bufpl[30];

	switch (Phase) {
	case 0:
		return;
	case 1:
	case 2:
		break;
	case 3:
		if ((firstindex = findplace(name,n1,n2)) == RES_ERROR)
			return;
		CHECK_A(n1,f1,l1,Arrplace[firstindex].n1)
		CHECK_A(n2,f2,l2,Arrplace[firstindex].n2)
		for (i1 = f1; i1 <= l1; i1++) {
		for (i2 = f2; i2 <= l2; i2++) {
			index = firstindex + i2 + i1 * Arrplace[firstindex].n2;
			if (init < 0 || init > MAX_TOKEN) {
				pr_place(bufpl,&Arrplace[index]);
				fprintf(stderr,"ERROR: illegal initial number of tokens for place '%s' (%d)\n",bufpl,init);
				Retval = RES_ERROR;
				return;
			}
			if (Arrplace[index].initial != 0) {
				pr_place(bufpl,&Arrplace[index]);
				fprintf(stderr,"ERROR: initial number of tokens for place '%s' already set\n",bufpl);
				Retval = RES_ERROR;
				return;
			}
			Arrplace[index].initial = init;
		}
		}
	}
}

void	Ctrans(name,n1,n2)
char	*name;		/* the name of the transition */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
Define transitions "name.0.0".."name.n1-1.n2-1" with initial priority 0
and enabling function NULL.
*/

{
	char	*aux;
	int	i1,i2;

	switch (Phase) {
	case 0:
		return;
	case 1:
		/* check if name is already present */
		if (insert_symb(name,SYM_TRANS,Ntrans) == RES_ERROR) {
			fprintf(stderr,"ERROR: illegal transition name ('%s')\n",name);
			Retval = RES_ERROR;
			return;
		}
		/* check size of transition array */
		if (n1 <= 0 || n2 <= 0 || n1 > MAX_ARRAY || n2 > MAX_ARRAY) {
			fprintf(stderr,"ERROR: invalid array size for transition '%s'\n",name);
			Retval = RES_ERROR;
			return;
		}
		Ntrans += (n1 * n2);
		break;
	case 2:
		aux = Arrtrans[Ntrans].name;
		for (i1 = 0; i1 < n1; i1++)
			for (i2 = 0; i2 < n2; i2++) {
				Arrtrans[Ntrans].name = aux;
				Arrtrans[Ntrans].i1 = i1;
				Arrtrans[Ntrans].i2 = i2;
				Arrtrans[Ntrans].n1 = n1;
				Arrtrans[Ntrans].n2 = n2;
				Arrtrans[Ntrans].type = DIS_NOVAL;
				Arrtrans[Ntrans].prio = 0;
				Arrtrans[Ntrans].enabl = NULL;
				Arrtrans[Ntrans].num = Ntrans;
				Ntrans++;
			}
		break;
	case 3:
		break;
	}
}

void	Cpriority(name,n1,n2,prio)
char	*name;		/* the name of the transition */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */
int	prio;		/* priority of the transition */

/*************************************************************************
Define the priority for transition "name.n1.n2".
*/

{
	char	buftr[30];
	int	firstindex,index,i1,i2,f1,f2,l1,l2;

	switch (Phase) {
	case 0:
		return;
	case 1:
	case 2:
		break;
	case 3:
		if ((firstindex = findtrans(name,n1,n2)) == RES_ERROR)
			return;
		CHECK_A(n1,f1,l1,Arrtrans[firstindex].n1)
		CHECK_A(n2,f2,l2,Arrtrans[firstindex].n2)
		for (i1 = f1; i1 <= l1; i1++) {
		for (i2 = f2; i2 <= l2; i2++) {
			index = firstindex + i2 + i1 * Arrtrans[firstindex].n2;
			/* check legality of priority */
			if (prio < 0 || prio > MAX_PRIO) {
				pr_trans(buftr,&Arrtrans[index]);
				fprintf(stderr,"ERROR: invalid priority for transition '%s'\n",buftr);
				Retval = RES_ERROR;
				return;
			}
			Arrtrans[index].prio = prio;
		}
		}
		break;
	}
}

void	Cenabling(name,n1,n2,efunc)
char	*name;		/* the name of the transition */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */
int	(*efunc)();	/* marking dependent enabling function of the transition */

/*************************************************************************
Define the enabling function for transition "name.n1.n2".
*/

{
	int	firstindex,index,i1,i2,f1,f2,l1,l2;

	switch (Phase) {
	case 0:
		return;
	case 1:
	case 2:
		break;
	case 3:
		if ((firstindex = findtrans(name,n1,n2)) == RES_ERROR)
			return;
		CHECK_A(n1,f1,l1,Arrtrans[firstindex].n1)
		CHECK_A(n2,f2,l2,Arrtrans[firstindex].n2)
		for (i1 = f1; i1 <= l1; i1++) {
		for (i2 = f2; i2 <= l2; i2++) {
			index = firstindex + i2 + i1 * Arrtrans[firstindex].n2;
			Arrtrans[index].enabl = efunc;
		}
		}
		break;
	}
}

/*VARARGS6*/
void	Cdistr(type,name,n1,n2,dep,val,dval,place,p1,p2,fun,dfun)
int	type;		/* DIS_x */
char	*name;		/* the name of the transition */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */
int	dep;		/* DEP_NO, DEP_PL, or DEP_MK */
double	val;		/* value, if DEP_NO or DEP_PL */
double	dval;		/* derivative value, if DEP_NO or DEP_PL */
char	*place;		/* place, if DEP_PL */
int	p1;		/* index on 1st dimension for the place */
int	p2;		/* index on 2nd dimension for the place */
double	(*fun) ();	/* marking dependent function, if DEP_MK */
double	(*dfun) ();	/* marking dependent function, if DEP_MK */

/*************************************************************************
Define the distributional parameters for transition "name.n1.n2".
If "type" == DIS_IMM: immediate transition.
If "type" == DIS_EXP: timed transition.
If "dep" == DEP_NO: probability or rate is "val".
If "dep" == DEP_PL: probability or rate is "val" * mark("place").
If "dep" == DEP_MK: probability or rate is "fun" evaluated in current marking.
*/

{
	char	buftr[30];
	int	firstindex,index,i1,i2,f1,f2,l1,l2,npl;

	switch (Phase) {
	case 0:
		return;
	case 1:
	case 2:
		break;
	case 3:
		if ((firstindex = findtrans(name,n1,n2)) == RES_ERROR)
			return;
		if (dep != DEP_PL) {
			npl = -1;
		} else {
			if ((npl = findplace(place,p1,p2)) == RES_ERROR)
				return;
			npl += p1 * Arrplace[npl].n2 + p2;
		}
		CHECK_A(n1,f1,l1,Arrtrans[firstindex].n1)
		CHECK_A(n2,f2,l2,Arrtrans[firstindex].n2)
		for (i1 = f1; i1 <= l1; i1++) {
		for (i2 = f2; i2 <= l2; i2++) {
			index = firstindex + i2 + i1 * Arrtrans[firstindex].n2;
			if (Arrtrans[index].type != DIS_NOVAL) {
				pr_trans(buftr,&Arrtrans[index]);
				fprintf(stderr,"ERROR: distribution for transition '%s' already set\n",buftr);
				Retval = RES_ERROR;
				return;
			}
			Arrtrans[index].type = type;
			Arrtrans[index].dep = dep;
			Arrtrans[index].pl = npl;
			Arrtrans[index].val = val;
			Arrtrans[index].dval = dval;
			Arrtrans[index].fun = fun;
			Arrtrans[index].dfun = dfun;
		}
		}
		break;
	}
}

/*VARARGS8*/
void	Carc(type,tr,t1,t2,pl,p1,p2,mult,func)
int	type;		/* ARC_x */
char	*tr;		/* name of the transition */
char	*pl;		/* name of the place */
int	mult;		/* constant multeplicity of the arc, or VARIABLE */
int	(*func)();	/* marking dependent multiplicity of the arc */

/*************************************************************************
Defines an arc with constant multeplicity "mult", or with variable multiplicity
func, if "mult" == VARIABLE.
If "type" == ARC_I:
	input arc from place "pl.p1.p2" to transition "tr.t1.t2".
If "type" == ARC_O:
	output arc from transition "tr.t1.t2" to place "pl.p1.p2".
If "type" == ARC_H:
	inhibitor arc from place "pl.p1.p2" to transition "tr.t1.t2".
*/

{
	char	buftr[30],bufpl[30];
	int	firstnpl,firstntr,npl,ntr;
	int	ip1,ip2,it1,it2,fp1,fp2,lp1,lp2,ft1,ft2,lt1,lt2;

	switch (Phase) {
	case 0:
		return;
	case 1:
	case 2:
		break;
	case 3:
		if ((firstnpl = findplace(pl,p1,p2)) == RES_ERROR)
			return;
		if ((firstntr = findtrans(tr,t1,t2)) == RES_ERROR)
			return;
		CHECK_A(p1,fp1,lp1,Arrplace[firstnpl].n1)
		CHECK_A(p2,fp2,lp2,Arrplace[firstnpl].n2)
		CHECK_A(t1,ft1,lt1,Arrtrans[firstntr].n1)
		CHECK_A(t2,ft2,lt2,Arrtrans[firstntr].n2)
		for (ip1 = fp1; ip1 <= lp1; ip1++) {
		for (ip2 = fp2; ip2 <= lp2; ip2++) {
			npl = firstnpl + ip2 + ip1 * Arrplace[firstnpl].n2;
		for (it1 = ft1; it1 <= lt1; it1++) {
		for (it2 = ft2; it2 <= lt2; it2++) {
			ntr = firstntr + it2 + it1 * Arrtrans[firstntr].n2;
			if (mult != VARIABLE && (mult <= 0 || mult > MAX_TOKEN)) {
				pr_trans(buftr,&Arrtrans[ntr]);
				pr_place(bufpl,&Arrplace[npl]);
				fprintf(stderr,"ERROR: illegal multeplicity (%d) in %carc ('%s','%s')\n",
					mult,type,buftr,bufpl);
				Retval = RES_ERROR;
				return;
			}
			switch (type) {
			case ARC_I:
				if (Arrtrans[ntr].inp[npl] != 0)
					break;
				Arrtrans[ntr].inp[npl] = mult;
				if (mult == VARIABLE) {
					if (Arrtrans[ntr].vin == NULL)
						Arrtrans[ntr].vin = (VA) new_mem((unsigned int)(sizeof(VA)*Nplace));
					Arrtrans[ntr].vin[npl] = func;
				}
				continue;
			case ARC_O:
				if (Arrtrans[ntr].out[npl] != 0)
					break;
				Arrtrans[ntr].out[npl] = mult;
				if (mult == VARIABLE) {
					if (Arrtrans[ntr].vou == NULL)
						Arrtrans[ntr].vou = (VA) new_mem((unsigned int)(sizeof(VA)*Nplace));
					Arrtrans[ntr].vou[npl] = func;
				}
				continue;
			case ARC_H:
				if (Arrtrans[ntr].inh == NULL)
					Arrtrans[ntr].inh = (PP *) new_mem((unsigned int)(sizeof(PP)*Nplace));
				if (Arrtrans[ntr].inh[npl] != 0)
					break;
				Arrtrans[ntr].inh[npl] = mult;
				if (mult == VARIABLE) {
					if (Arrtrans[ntr].vih == NULL)
						Arrtrans[ntr].vih = (VA) new_mem((unsigned int)(sizeof(VA)*Nplace));
					Arrtrans[ntr].vih[npl] = func;
				}
				continue;
			}
			pr_trans(buftr,&Arrtrans[ntr]);
			pr_place(bufpl,&Arrplace[npl]);
			fprintf(stderr,"ERROR: more than one %carc ('%s','%s')\n",type,buftr,bufpl);
			Retval = RES_ERROR;
			return;
		}
		}
		}
		}
	}
}

int install()

/*************************************************************************
It installs the PN into the data structure.
RETURNS:
RES_NOERR or RES_ERROR
*/

{
	char	buftr[30];
	char	*symb_chars =
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	PP	*auxpp;
	int	i,j;

	for (i = 0; i < MAX_CHAR; i++)
		Charval[i] = 0;
	for (j = 1, i = 0; symb_chars[i] != '\0'; i++)
		Charval[symb_chars[i]] = j++;
	for (i = 0; i < HSH_SIZE; i++)
		Symtable[i].type = SYM_NOVAL;
	Phase = 0;	/* get the parameters */
	parameters();
	Phase = 1;	/* count the  number of places and transitions */
	Nplace = Ntrans = 0;
	net();
	if (Retval != RES_NOERR)
		return(Retval);
	Phase = 2;	/* consider place and trans */
	Arrplace = (PL	*) new_mem((unsigned int)(sizeof(PL) * Nplace));
	Arrtrans = (TR	*) new_mem((unsigned int)(sizeof(TR) * Ntrans));
	auxpp = (PP *) new_mem((unsigned int)(sizeof(PP)*Nplace*Ntrans*2));
	for (i = 0; i < Ntrans; i++) {
		Arrtrans[i].inp = &auxpp[2 * Nplace * i];
		Arrtrans[i].out = &auxpp[2 * Nplace * i + Nplace];
	}
	for (i = 0; i < HSH_SIZE; i++)
		switch (Symtable[i].type) {
		case SYM_PLACE:
			Arrplace[Symtable[i].numb].name = Symtable[i].symb;
			break;
		case SYM_TRANS:
			Arrtrans[Symtable[i].numb].name = Symtable[i].symb;
			break;
		case SYM_PARAM:
			fprintf(stderr,"WHAT?\n");
			exit(10);
		case SYM_NOVAL:
			break;
		}
	Nplace = Ntrans = 0;
	net();
	if (Retval != RES_NOERR)
		return(Retval);
	Phase = 3;
	net();
	for (i = 0; i < Ntrans; i++)
		if (Arrtrans[i].type == DIS_NOVAL) {
			pr_trans(buftr,&Arrtrans[i]);
			fprintf(stderr,"ERROR: distribution type for '%s' not specified\n",buftr);
			Retval = RES_ERROR;
		}
	if (Retval != RES_NOERR)
		return(Retval);
	Phase = 0;
	return(RES_NOERR);
}

double	input(descr)
char	*descr;

/*************************************************************************
User available function to input a double parameter.
The value is recorded in ".out" as
		INPUT: "descr" = <value>;
RETURNS:
a value (double)
*/

{
extern	FILE	*Outfile;
double	parm;

	if (Phase != 0) {
		fprintf(stderr,"ERROR: 'input' must be called in 'parameters'\n");
		Retval = RES_ERROR;
		return(RES_ERROR);
	}
	for (;;) {
		fprintf(stderr,"INPUT \"%s\">\n",descr);
		if (scanf("%lf",&parm) != 1) {
			while(getchar() != '\n');
			fprintf(stderr,"ERROR, RETRY ");
		} else
			break;
	}
	fprintf(Outfile,"INPUT: %s = %lg\n",descr,parm);
	return(parm);
}

int	Cmark(pl,n1,n2)
char	*pl;		/* place name */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
User available function to express the number of tokens in place "pl.n1.n2".
RETURNS:
the number of tokens for place "pl.n1.n2" in marking "Currmark"
*/

{
extern	ME	*Currmark;
	int	tok,val1,val2,flag1,flag2;
	int	firstindex,f1,l1,f2,l2,i1,i2;
	
	if ((firstindex = findplace(pl,n1,n2)) == RES_ERROR)
		return(0);
	CHECK_SMM(n1,f1,l1,Arrplace[firstindex].n1,flag1)
	CHECK_SMM(n2,f2,l2,Arrplace[firstindex].n2,flag2)
	SWITCHFLG(f1,l1,i1,val1,val2,SWITCHFLG(f2,l2,i2,val2,tok,tok=Currmark->mk[firstindex+i2+i1*Arrplace[firstindex].n2];,flag2,MAX_TOKEN),flag1,MAX_TOKEN)
	return(val1);
}

int internal_enabled(tr)
TR	*tr;

/*************************************************************************
Internal function to test the enabling of the transition pointed by "tr".
RETURNS:
1 if "tr" is enabled in marking "Currmark", 0 otherwise
*/

{
extern	ME	*Currmark;
	int	i,aux;

	/* check inhibitor arcs */
	if (tr->inh != NULL) {
		for (i = 0; i < Nplace; i++) {
			if (tr->inh[i] == VARIABLE) {
				aux = (*(tr->vih[i]))(tr->i1,tr->i2);
				if (aux > 0 && Currmark->mk[i] >= aux)
					return(0);
			} else {
				if (tr->inh[i] > 0 && Currmark->mk[i] >= tr->inh[i])
					return(0);
			}
		}
	}
	/* check constant input arcs */
	for (i = 0; i < Nplace; i++) {
		if (tr->inp[i] == VARIABLE) {
			if ((*(tr->vin[i]))(tr->i1,tr->i2) > Currmark->mk[i])
				return(0);
		} else {
			if (tr->inp[i] > Currmark->mk[i])
			return(0);
		}
	}
	/* check whether the enabling function is 0 */
	if (tr->enabl != NULL && !((*tr->enabl)(tr->i1,tr->i2)) )
		return(0);
	return(1);
}

double	internal_rate(tr)
TR	*tr;

/*************************************************************************
Internal function to return the rate of the transition pointed by "tr".
RETURNS:
the rate of "tr" in marking "Currmark", if enabled, 0 otherwise
*/

{
extern	ME	*Currmark;

	if (tr->type != DIS_EXP) {
		fprintf(stderr,"ERROR: rate is not defined for immediate transitions\n");
		Retval = RES_ERROR;
		return(0.0);
	}
	if (! internal_enabled(tr))
		return(0.0);
	switch (tr->dep) {
	case DEP_NO:
		return(tr->val);
	case DEP_PL:
		return(tr->val * Currmark->mk[tr->pl]);
	case DEP_MK:
		return((*tr->fun)(tr->i1,tr->i2));
	default:
		exit(ER_CODE);
		return(0.0);		/* avoids lint complaints */
	}
}

int	Cenabled(tr,n1,n2)
char	*tr;		/* the transition name */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
User available function to test the enabling of transition "tr.n1.n2".
RETURNS:
1 if "tr.n1.n2" is enabled in marking "Currmark", 0 otherwise
*/
{
	int	firstindex,flag1,flag2,f1,f2,l1,l2,i1,i2,val1,val2,ena;
	
	if ((firstindex = findtrans(tr,n1,n2)) == RES_ERROR)
		return(0);
	CHECK_SMM(n1,f1,l1,Arrtrans[firstindex].n1,flag1)
	CHECK_SMM(n2,f2,l2,Arrtrans[firstindex].n2,flag2)
	SWITCHFLG(f1,l1,i1,val1,val2,SWITCHFLG(f2,l2,i2,val2,ena,ena=internal_enabled(&Arrtrans[firstindex+i2+i1*Arrtrans[firstindex].n2]);,flag2,MAX_TOKEN),flag1,MAX_TOKEN)
	return(val1);
}

double	Crate(tr,n1,n2)
char	*tr;		/* the transition name */
int	n1;		/* index on 1st dimension */
int	n2;		/* index on 2nd dimension */

/*************************************************************************
User available function to obtain the rate of transition "tr.n1.n2".
RETURNS:
the rate of "tr.n1.n2" if is enabled in marking "Currmark", 0 otherwise
*/

{
	int	firstindex,flag1,flag2,f1,f2,l1,l2,i1,i2;
	double	val1,val2,rt;
	
	if ((firstindex = findtrans(tr,n1,n2)) == RES_ERROR)
		return(0);
	CHECK_SMM(n1,f1,l1,Arrtrans[firstindex].n1,flag1)
	CHECK_SMM(n2,f2,l2,Arrtrans[firstindex].n2,flag2)
	SWITCHFLG(f1,l1,i1,val1,val2,SWITCHFLG(f2,l2,i2,val2,rt,rt=internal_rate(&Arrtrans[firstindex+i2+i1*Arrtrans[firstindex].n2]);,flag2,MAX_FLOAT),flag1,MAX_FLOAT)
	return(val1);
}

/*VARARGS0*/
double	expected(func)
double	(*func)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to compute the expected value of the marking dependent
expression represented by "func".
*/

{
	double	val = 0.0;
	
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if (Currmark->index < 0)
			continue;
		val += (*func)() * CProb[Currmark->index];
	}
	return(val);
}

/*VARARGS0*/
double	der_expected(func,derfunc)
double	(*func)();	/* the function whose expected value is computed */
double	(*derfunc)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to compute the expected value of the marking dependent
expression represented by "func".
*/

{
	double	val = 0.0;

	if ( Method_sensitivity == VAL_NO )
		return(0.0);
	
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if (Currmark->index < 0)
			continue;
		val += (*func)() * DerCProb[Currmark->index]
			+ (*derfunc)() * CProb[Currmark->index];
	}
	return(val);
}

/*VARARGS0*/

/*VARARGS0*/
double	cum_expected(func)
double	(*func)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to compute the expected value of the marking dependent
expression represented by "func".
*/

{
	double	val = 0.0;

	if ( Method_transient == VAL_NO )
		return(0.0);
	
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if (Currmark->index < 0)
			continue;
		val += (*func)() * Cumprob[Currmark->index];
	}
	return(val);
}

/*VARARGS0*/
double	der_cum_expected(func,derfunc)
double	(*func)();	/* the function whose expected value is computed */
double	(*derfunc)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to compute the expected value of the marking dependent
expression represented by "func".
*/

{
	double	val = 0.0;

	if ( Method_sensitivity == VAL_NO )
		return(0.0);

	if ( Method_transient == VAL_NO )
		return(0.0);
	
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if (Currmark->index < 0)
			continue;
		val += (*func)() * DerCumprob[Currmark->index]
		       + (*derfunc)() * Cumprob[Currmark->index];
	}
	return(val);
}

double	accumulated(func)
double	(*func)();	/* the function whose accumulated value is computed */

/*************************************************************************
User available function to compute the accumulated value of the marking
dependent expression represented by "func".
*/

{
static	double	*reward = NULL;
static	double	*visits = NULL;
	double	val;
	int	abs;
	int	i;
	int	iter,method;
	double	prec;
	
	if (reward == NULL) {
		reward = (double *)(new_mem((unsigned int)(sizeof(double) *
		Nstates)));
		reward = &reward[-Find];
		visits = (double *)(new_mem((unsigned int)(sizeof(double) *
		Nstates)));
		visits = &visits[-Find];
	}
	if (Prob0flag == VAL_NO) {
		fprintf(stderr,"ERROR: define initial probabilities first\n");
		exit(ER_CODE);
	}
	Prob0flag = VAL_NO;
	for (i = Find; i < Ntottan; i++)
		visits[i] = 0.0;
	abs = VAL_NO;
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if (Currmark->index < 0 && IOpt[IOP_MC].val == VAL_CTMC)
			continue;
		val = (*func)();
		if (val == 0.0)
			abs = VAL_YES;
		reward[Currmark->index] = val;
	}
	if (abs == VAL_NO) {
		fprintf(stderr,"ERROR: 'accumulated' requires at least one absorbing state\n");
		exit(ER_CODE);
	}
	iter = IOpt[IOP_ITERATIONS].val;
	prec = FOpt[FOP_PRECISION].val;
	method = IOpt[IOP_METHOD].val;
	modsor(IP,JA,A,reward,Find,Ntottan-1,visits,Prob0,&prec,&iter,&method);
	val = 0.0;
	if (IOpt[IOP_MC].val == VAL_DTMC) {
		for (i = 0; i < Ntottan; i++)
			val += visits[i] * Holding[i] * reward[i];
	} else {
		for (i = 0; i < Ntottan; i++)
			val += visits[i] * reward[i];
	}
	if (prec > FOpt[FOP_PRECISION].val) {
		fprintf(stderr,"ERROR/WARNING: required precision not reached in 'accumulated'\n");
	}
	return(val);
}

void	hold_cond(cond,times)
int	(*cond)();
double	times[2];

/*************************************************************************
Computes into times[VAL_YES] the expected time a condition holds and
into times[VAL_NO] the expected time a condition does not hold.
*/

{
extern	void	col2row();
static	char	*mk_cond = NULL;
	int	j,i,c;
	double	v_yes = 0.0;
	double	p_yes = 0.0;
	double	v_no = 0.0;
	double	p_no = 0.0;
	
	if (mk_cond == NULL) {			/* allocate it */
		mk_cond = (char *)(new_mem((unsigned int) (sizeof(char)*(Nvan+Ntottan))));
		mk_cond = &mk_cond[Nvan];
	}
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		if ((c = (*cond)()) != VAL_NO && c != VAL_YES) {
			fprintf(stderr,"ERROR: condition function does not evaluate to VAL_NO or VAL_YES\n");
			exit(ER_CODE);
		}
		mk_cond[Currmark->index] = c;	/* fill 0 or 1 for mk_cond[] */
	}
	if (IOpt[IOP_MC].val == VAL_CTMC) {
		for (i = Find; i < Ntottan; i++) {
			if (mk_cond[i] == VAL_YES) {
				p_yes += CProb[i];
				for (j = IP[i]; j < IP[i+1]; j++)
					if (mk_cond[JA[j]] == VAL_NO) /*N->Y*/
						v_no += CProb[JA[j]] * A[j];
			} else {
				p_no += CProb[i];
				for (j = IP[i]; j < IP[i+1]; j++)
					if (mk_cond[JA[j]] == VAL_YES) /*Y->N*/
						v_yes += CProb[JA[j]] * A[j];
			}
		}
	} else {
		for (i = Find; i < Ntottan; i++) {
			if (mk_cond[i] == VAL_YES) {
				if (i >= 0)
					p_yes += DProb[i] * Holding[i];
				for (j = IP[i]; j < IP[i+1]; j++)
					if (mk_cond[JA[j]] == VAL_NO) /*N->Y*/
						v_no += DProb[JA[j]] * A[j];
			} else {
				if (i >= 0)
				p_no += DProb[i] * Holding[i];
				for (j = IP[i]; j < IP[i+1]; j++)
					if (mk_cond[JA[j]] == VAL_YES) /*Y->N*/
						v_yes += DProb[JA[j]] * A[j];
			}
		}
	}
	if (p_yes == 0.0) {
		fprintf(stderr,"ERROR: condition function never holds\n");
		exit(ER_CODE);
	}
	if (p_no == 0.0) {
		fprintf(stderr,"ERROR: condition function always holds\n");
		exit(ER_CODE);
	}
	times[VAL_NO] = p_no / v_no;
	times[VAL_YES] = p_yes / v_yes;
} /* hold_cond */

void	pr_std_average()

/*************************************************************************
Print the standard average measures.
*/

{
extern	void	col2row();
	double	*prpl;
	double	*avpl;
	double	*prtr;
	double	*avtr;
	double	prabs = 0.0;
	ME	*auxme;
	EE	*auxee;
	int	i;
	char	aux[200];

	/* allocate the arrays */
	prpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	avpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	prtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	avtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));

	for (i = 0; i < Nplace; i++)
		prpl[i] = avpl[i] = 0.0;
	for (i = 0; i < Ntrans; i++)
		prtr[i] = avtr[i] = 0.0;

	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->index < 0)		/* vanishing marking, ignore */
			continue;
		for (i = 0; i < Nplace; i++) {
			if (auxme->mk[i] == 0)
				continue;
			prpl[i] += CProb[auxme->index];
			avpl[i] += CProb[auxme->index] * auxme->mk[i];
		}
	}
	if (Direction == VAL_TOFROM) {
		col2row(-Nvan,Ntottan-1);
		Direction = VAL_FROMTO;
	}

	for (i = 0; i < Ntan; i++) {
		for (auxee = Qrow[i]; auxee != NULL; auxee = auxee->row.pt) {
			prtr[auxee->tr] += CProb[i];
			avtr[auxee->tr] += CProb[i] * (double)(auxee->value);
		}
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		for (; i < Ntottan; i++)
			prabs += CProb[i];
		
	fprintf(Outfile,"AVERAGE:\n==========================================================================\n");
	fprintf(Outfile,"    PLACE                        Pr[nonempty]          Av[tokens]\n");
	for (i = 0; i < Nplace; i++) {
		pr_place(aux,&Arrplace[i]);
		fprintf(Outfile,"%3d: %-28s %18.12le    %18.12le\n",i,aux,prpl[i],avpl[i]);
	}
	fprintf(Outfile,"    TRANSITION                   Pr[enabled]           Av[throughput]\n");
	for (i = 0; i < Ntrans; i++) {
		if (Arrtrans[i].type == DIS_IMM)
			continue;
		pr_trans(aux,&Arrtrans[i]);
		fprintf(Outfile,"%3d: %-28s %18.12le    %18.12le\n",i,aux,prtr[i],avtr[i]);
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		fprintf(Outfile,"     %-28s %18.12le    %18.12le\n","(return to M0)",prabs,(double)(prabs*(double)(FOpt[FOP_ABS_RET_M0].val)));
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_std_average_der()

/*************************************************************************
Print the standard average measures.
*/

{
extern	void	col2row();
	double	*prpl;
	double	*avpl;
	double	*prtr;
	double	*avtr;
	double	prabs = 0.0;
	ME	*auxme;
	EE	*auxee;
	int	i;
	char	aux[200];

	if ( Method_sensitivity == VAL_NO )
		return;

	/* allocate the arrays */
	prpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	avpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	prtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	avtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));

	for (i = 0; i < Nplace; i++)
		prpl[i] = avpl[i] = 0.0;
	for (i = 0; i < Ntrans; i++)
		prtr[i] = avtr[i] = 0.0;

	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->index < 0)		/* vanishing marking, ignore */
			continue;
		for (i = 0; i < Nplace; i++) {
			if (auxme->mk[i] == 0)
				continue;
			prpl[i] += DerCProb[auxme->index];
			avpl[i] += DerCProb[auxme->index] * auxme->mk[i];
		}
	}
	if (Direction == VAL_TOFROM) {
		col2row(-Nvan,Ntottan-1);
		Direction = VAL_FROMTO;
	}

	for (i = 0; i < Ntan; i++) {
		for (auxee = Qrow[i]; auxee != NULL; auxee = auxee->row.pt) {
			prtr[auxee->tr] += DerCProb[i];
			avtr[auxee->tr] += DerCProb[i] * (double)(auxee->value);
		}
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		for (; i < Ntottan; i++)
			prabs += DerCProb[i];
		
	fprintf(Outfile,"DERIVATIVES:\n==========================================================================\n");
	fprintf(Outfile,"    PLACE                        Pr[nonempty]          Av[tokens]\n");
	for (i = 0; i < Nplace; i++) {
		pr_place(aux,&Arrplace[i]);
		fprintf(Outfile,"%3d: %-28s %18.12le    %18.12le\n",i,aux,prpl[i],avpl[i]);
	}
	fprintf(Outfile,"    TRANSITION                   Pr[enabled]           Av[throughput]\n");
	for (i = 0; i < Ntrans; i++) {
		if (Arrtrans[i].type == DIS_IMM)
			continue;
		pr_trans(aux,&Arrtrans[i]);
		fprintf(Outfile,"%3d: %-28s %18.12le    %18.12le\n",i,aux,prtr[i],avtr[i]);
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		fprintf(Outfile,"     %-28s %18.12le    %18.12le\n","(return to M0)",prabs,(double)(prabs*(double)(FOpt[FOP_ABS_RET_M0].val)));
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_std_cum_average()

/*************************************************************************
Print the standard cumulative average measures.
*/

{
extern	void	col2row();
	double	*prpl;
	double	*avpl;
	double	*prtr;
	double	*avtr;
	double	prabs = 0.0;
	ME	*auxme;
	EE	*auxee;
	int	i;
	char	aux[200];

	if ( Method_transient == VAL_NO )
	    return;

	/* allocate the arrays */
	prpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	avpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	prtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	avtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	for (i = 0; i < Nplace; i++)
		prpl[i] = avpl[i] = 0.0;
	for (i = 0; i < Ntrans; i++)
		prtr[i] = avtr[i] = 0.0;

	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->index < 0)		/* vanishing marking, ignore */
			continue;
		for (i = 0; i < Nplace; i++) {
			if (auxme->mk[i] == 0)
				continue;
			prpl[i] += Cumprob[auxme->index];
			avpl[i] += Cumprob[auxme->index] * auxme->mk[i];
		}
	}
	if (Direction == VAL_TOFROM) {
		col2row(-Nvan,Ntottan-1);
		Direction = VAL_FROMTO;
	}
	for (i = 0; i < Ntan; i++)
		for (auxee = Qrow[i]; auxee != NULL; auxee = auxee->row.pt) {
			prtr[auxee->tr] += Cumprob[i];
			avtr[auxee->tr] += Cumprob[i] * (double)(auxee->value);
		}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs)
		for (; i < Ntottan; i++)
			prabs += Cumprob[i];
		
	fprintf(Outfile,"CUMULATIVE AVERAGE:\n==========================================================================\n");
	fprintf(Outfile,"    PLACE                        Pr[nonempty]          Av[tokens]\n");
	for (i = 0; i < Nplace; i++) {
		pr_place(aux,&Arrplace[i]);
		fprintf(Outfile,"%3d: %-28s %18.12e    %18.12e\n",i,aux,prpl[i],avpl[i]);
	}
	fprintf(Outfile,"    TRANSITION                   Pr[enabled]           Av[throughput]\n");
	for (i = 0; i < Ntrans; i++) {
		if (Arrtrans[i].type == DIS_IMM)
			continue;
		pr_trans(aux,&Arrtrans[i]);
		fprintf(Outfile,"%3d: %-28s %18.12e    %18.12e\n",i,aux,prtr[i],avtr[i]);
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		fprintf(Outfile,"     %-28s %18.12e    %18.12e\n","(return to M0)",prabs,(double)(prabs*(double)(FOpt[FOP_ABS_RET_M0].val)));
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_std_cum_average_der()

/*************************************************************************
Print the standard cumulative average measures.
*/

{
extern	void	col2row();
	double	*prpl;
	double	*avpl;
	double	*prtr;
	double	*avtr;
	double	prabs = 0.0;
	ME	*auxme;
	EE	*auxee;
	int	i;
	char	aux[200];

	if ( Method_transient == VAL_NO )
	    return;
	if ( Method_sensitivity == VAL_NO )
	    return;

	/* allocate the arrays */
	prpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	avpl = (double *) new_mem((unsigned int) (sizeof(double)*Nplace));
	prtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	avtr = (double *) new_mem((unsigned int) (sizeof(double)*Ntrans));
	for (i = 0; i < Nplace; i++)
		prpl[i] = avpl[i] = 0.0;
	for (i = 0; i < Ntrans; i++)
		prtr[i] = avtr[i] = 0.0;

	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->index < 0)		/* vanishing marking, ignore */
			continue;
		for (i = 0; i < Nplace; i++) {
			if (auxme->mk[i] == 0)
				continue;
			prpl[i] += DerCumprob[auxme->index];
			avpl[i] += DerCumprob[auxme->index] * auxme->mk[i];
		}
	}
	if (Direction == VAL_TOFROM) {
		col2row(-Nvan,Ntottan-1);
		Direction = VAL_FROMTO;
	}
	for (i = 0; i < Ntan; i++)
		for (auxee = Qrow[i]; auxee != NULL; auxee = auxee->row.pt) {
			prtr[auxee->tr] += DerCumprob[i];
			avtr[auxee->tr] += DerCumprob[i] * (double)(auxee->value);
		}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs)
		for (; i < Ntottan; i++)
			prabs += DerCumprob[i];
		
	fprintf(Outfile,"CUMULATIVE DERIVATIVE:\n==========================================================================\n");
	fprintf(Outfile,"    PLACE                        Pr[nonempty]          Av[tokens]\n");
	for (i = 0; i < Nplace; i++) {
		pr_place(aux,&Arrplace[i]);
		fprintf(Outfile,"%3d: %-28s %18.12e    %18.12e\n",i,aux,prpl[i],avpl[i]);
	}
	fprintf(Outfile,"    TRANSITION                   Pr[enabled]           Av[throughput]\n");
	for (i = 0; i < Ntrans; i++) {
		if (Arrtrans[i].type == DIS_IMM)
			continue;
		pr_trans(aux,&Arrtrans[i]);
		fprintf(Outfile,"%3d: %-28s %18.12e    %18.12e\n",i,aux,prtr[i],avtr[i]);
	}
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		fprintf(Outfile,"     %-28s %18.12e    %18.12e\n","(return to M0)",prabs,(double)(prabs*(double)(FOpt[FOP_ABS_RET_M0].val)));
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_net_info()

/*************************************************************************
Print the information on the SPN size on the "Outfile".
*/

{
	int	i,j,ninp,nout,ninh,nvinp,nvout,nvinh,nimm,nexp;

	fprintf(Outfile,"NET:\n");
	fprintf(Outfile,"==========================================================================\n");
	fprintf(Outfile,"\tplaces: %d\n",Nplace);
	nimm = nexp = 0;
	nvinp = nvout = nvinh = ninp = nout = ninh = 0;
	for (i = 0; i < Ntrans; i++) {
		switch (Arrtrans[i].type) {
			case DIS_EXP:	nexp++;	break;
			case DIS_IMM:	nimm++;	break;
		}
		for (j = 0; j < Nplace; j++) {
			if (Arrtrans[i].inp[j] == VARIABLE)
				nvinp++;
			else if (Arrtrans[i].inp[j] > 0)
				ninp++;
			if (Arrtrans[i].out[j] == VARIABLE)
				nvout++;
			else if (Arrtrans[i].out[j] > 0)
				nout++;
		}
		if (Arrtrans[i].inh != NULL) 
			for (j = 0; j < Nplace; j++) {
				if (Arrtrans[i].inh[j] == VARIABLE) {
					nvinh++;
                }
				else {
                    if (Arrtrans[i].inh[j] > 0) {
                        ninh++;
                    }
                }
            }
	}
	fprintf(Outfile,"\timmediate transitions: %d\n",nimm);
	fprintf(Outfile,"\ttimed transitions: %d\n",nexp);
	fprintf(Outfile,"\tconstant input arcs: %d\n",ninp);
	fprintf(Outfile,"\tconstant output arcs: %d\n",nout);
	fprintf(Outfile,"\tconstant inhibitor arcs: %d\n",ninh);
	fprintf(Outfile,"\tvariable input arcs: %d\n",nvinp);
	fprintf(Outfile,"\tvariable output arcs: %d\n",nvout);
	fprintf(Outfile,"\tvariable inhibitor arcs: %d\n",nvinh);
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_rg_info()

/*************************************************************************
Print the information on the reachability graph size on "Outfile".
*/

{
	fprintf(Outfile,"RG:\n");
	fprintf(Outfile,"==========================================================================\n");
	fprintf(Outfile,"\ttangible markings: %d",Ntottan);
	if (Nabs > 0)
		fprintf(Outfile," (%d absorbing)",Nabs);
	fprintf(Outfile,"\n");
	fprintf(Outfile,"\tvanishing markings: %d",Nvan);
	fprintf(Outfile,"\n");
	fprintf(Outfile,"\tmarking-to-marking transitions: %d\n",Narcs);
	fprintf(Outfile,"==========================================================================\n");
}

void	pr_mc_info()

/*************************************************************************
Print the information on the MC size on the "Outfile".
*/

{
	char	string;
	int	states;

	if (IOpt[IOP_MC].val == VAL_CTMC) {
		string = 'C';
		states = Ntottan;
	} else {
		string = 'D';
		states = Ntottan + Nvan;
	}
	fprintf(Outfile,"%cTMC:\n",string);
	fprintf(Outfile,"==========================================================================\n");
	fprintf(Outfile,"\tstates: %d",states);
	if (Nabs > 0)
		fprintf(Outfile," (%d absorbing)",Nabs);
	fprintf(Outfile,"\n");
	fprintf(Outfile,"\tnonzero entries: %d\n",Nentries);
	fprintf(Outfile,"\titerations: %d\n",Iterations);
	fprintf(Outfile,"\tprecision: %lg\n",Precision);
	fprintf(Outfile,"==========================================================================\n");
}

/*VARARGS1*/
void	pr_expected(string,func)
char	*string;	/* the string to be printed on output */
double	(*func)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to request the expected value of the marking dependent
expression represented by "func".
It prints on the ".out" file
		EXPECTED: "string" = <expected value>;
*/

{
	fprintf(Outfile,"EXPECTED: %s = %.12lg\n",string,expected(func));
}

/*VARARGS1*/
void	pr_sens_expected(string,func,derfunc)
char	*string;	/* the string to be printed on output */
double	(*func)();	/* the function whose expected value is computed */
double	(*derfunc)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to request the expected value of the marking dependent
expression represented by "func".
It prints on the ".out" file
		EXPECTED: "string" = <expected value>;
*/

{
	if ( Method_sensitivity == VAL_NO )
	   return;
	fprintf(Outfile,"EXPECTED: %s = %.12lg\n",string,expected(func));
	fprintf(Outfile,"Derivative: %s = %.12lg\n",
			string,der_expected(func,derfunc));
}

/*VARARGS1*/
void	pr_cum_expected(string,func)
char	*string;	/* the string to be printed on output */
double	(*func)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to request the expected value of the marking dependent
expression represented by "func".
It prints on the ".out" file
		EXPECTED: "string" = <expected value>;
*/

{
	if ( Method_transient == VAL_NO )
		return;
	fprintf(Outfile,"CUMULATIVE EXPECTED: %s = %.12g\n",
			       string,cum_expected(func));
}

/*VARARGS1*/
void	pr_time_avg_expected(string,func)
char	*string;	/* the string to be printed on output */
double	(*func)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to request the expected value of the marking dependent
expression represented by "func".
It prints on the ".out" file
		EXPECTED: "string" = <expected value>;
*/

{
	fprintf(Outfile,"TIME AVERAGE EXPECTED: %s = %.12g\n",
			       string, cum_expected(func)/Time );
}

/*VARARGS1*/
void	pr_sens_cum_expected(string,func,derfunc)
char	*string;	/* the string to be printed on output */
double	(*func)();	/* the function whose expected value is computed */
double	(*derfunc)();	/* the function whose expected value is computed */

/*************************************************************************
User available function to request the expected value of the marking dependent
expression represented by "func".
It prints on the ".out" file
		EXPECTED: "string" = <expected value>;
*/

{
	if ( Method_transient == VAL_NO )
		return;
	if ( Method_sensitivity == VAL_NO )
	   return;
	fprintf(Outfile,"CUMULATIVE EXPECTED: %s = %.12g\n",
			       string,cum_expected(func));
	fprintf(Outfile,"Derivative: %s = %.12lg\n",
			string,der_cum_expected(func,derfunc));
}

void	pr_accumulated(string,func)
char	*string;
double	(*func)();

/*************************************************************************
Print the accumulated reward up to absorption using func to define which
states are absorbing (func = 0) and the rewards of the other states (func != 0).
*/

{
	fprintf(Outfile,"ACCUMULATED: %s = %.12lg\n",string,accumulated(func));
}

void	pr_hold_cond(string,cond)
char	*string;	/* the string to be printed on output */
int	(*cond)();	/* the condition to be timed for ON/OFF */

/*************************************************************************
User available function to request the expected time ON/OFF for condition
represented by "cond".
It prints on the ".out" file
		TIME_YES: "string" = <expected value>;
		TIME_NO: "string" = <expected value>;
*/

{	double	aux[2];

	hold_cond(cond,aux);
	fprintf(Outfile,"TIME_YES: %s = %.12lg\n",string,aux[VAL_YES]);
	fprintf(Outfile,"TIME_NO: %s = %.12lg\n",string,aux[VAL_NO]);
}

void	Cset_prob0(scale,func)
int	scale;		/* VAL_YES or VAL_NO */
double	(*func)();	/* the function to define the initial prob */

/*************************************************************************
User available function to set the inital probability of marking m to
		pi0(m) = pi(m)func(m) / SUMi pi(i)func(i)
if scale == VAL_YES
		pi0(m) = func(m) / SUMi func(i)
if scale == VAL_NO
*/

{	int	i;
	double	val,sum;

	if (Prob0 == NULL) {
		Prob0 = (double *)(new_mem((unsigned int)(sizeof(double) *
		Nstates)));
		Prob0 = &Prob0[-Find];
	}
	for (i = Find; i < Ntottan; i++)
		Prob0[i] = 0.0;
	sum = 0.0;
	for (Currmark = Pcan; Currmark != NULL; Currmark = Currmark->pcan) {
		i = Currmark->index;
		if (scale == VAL_YES) {
			if (i < 0)
				continue;
			val = (*func)() * CProb[i];
		} else {
			if (i < 0 && IOpt[IOP_MC].val == VAL_CTMC) {
				val = (*func)();
				if (val > 1.0e-30) {
					fprintf(stderr,"ERROR: initial probability for vanishing markings must be zero if you use CTMC\n");
					exit(ER_CODE);
				} else {
					continue;
				}
			}
			val = (*func)();
		}
		if (val > 1.0e-30) {
			sum += val;
			Prob0[i] = val;
		}
	}
	if (sum == 0.0) {
		fprintf(stderr,"ERROR: 'set_prob' requires at least one positive probability\n");
		exit(ER_CODE);
	}
	for (i = Find; i < Ntottan; i++) {
		Prob0[i] /= sum;
	}
	Prob0flag = VAL_YES;
} /* Cset_prob0 */

/********************************************************
  Prints the time point at which the measures are computed, in the output file
  */
void print_time( time_value )
double time_value;
{
   fprintf(Outfile,"==========================================================================\n");
   fprintf(Outfile,"\n                         TIME :   %.12f\n\n", time_value );
   fprintf(Outfile,"==========================================================================\n");
}
