/*************************************************************************
FILE:	lex.h
*/

/*	definition related to the keywords for RG		*/
# define	R_T		0
# define	R_A		1
# define	R_V		2
# define	R_L		3
# define	R_nplace	4
# define	R_ntrans	5
# define	R_places	6
# define	R_transitions	7
# define	R_ntanmark	8
# define	R_nabsmark	9
# define	R_nvanmark	10
# define	R_nvanloop	11
# define	R_nentries	12
# define	R_reachset	13
# define	R_reachgraph	14

/***************************************************************/

/* Codes for the tokens returned by the lexical analyzer */
# define	NEWLINE		'\n'
# define	DIGITS		"0123456789"
# define	LETTERS	"_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
# define	ALPHA_SIZE	257	/* # of characters + 1 (for EOF) */
# define	MAXBUFFER	256	/* dimension of the char buffer */

# define	NO		0
# define	YES		1
# define	NONE		-1
# define	ASK		-2

/* Code groups for the characters, used to reduce entries in Nexstate */
# define	C_LETT		0	/* letter: '_','A',..,'Z','a',..,'z' */
# define	C_MINU		1	/* minus: '-' */
# define	C_PLUS		2	/* plus: '+' */
# define	C_DIGI		3	/* digit: '0',..,'9' */
# define	C_PERI		4	/* period: '.' */
# define	C_COLO		5	/* colon: ':' */
# define	C_EQUA		6	/* equal: '=' */
# define	C_SPAC		7	/* space: '\b','\t' */
# define	C_NEWL		8	/* newline: '\n' */
# define	C_SEMI		9	/* semicolon: ';' */
# define	C_REMA		10	/* remark: '#' */
# define	C_EOF		11	/* end of file: EOF */
# define	C_ILLE		12	/* illegal char, the remaining ones */

/* Codes for the tokens returned by the lexical analyzer */
# define	L_STRING	1021	/* a string not in the symbol table */
# define	L_EQUAL		1022	/* equal sign */
# define	L_SEMICOLON	1023	/* semicolon sign */
# define	L_COLON		1024	/* colon sign */
# define	L_NUMBER	1025	/* integer number */
# define	L_FLOAT		1026	/* float number */
# define	L_EOF		1027	/* end of file */

# ifdef LEXMAIN
# define	EX
# else
# define	EX extern
# endif

EX char	**Table;
/* It is used to look up the keywords read by the lexical analyzer.
   It is used by "look()". Different files have different keywords */

EX int	Nkeyword;
/* The number of keywords in Table */

EX char	*Gr;
/* It is used to group the chars for the lexical analyzer.
   It is initialized outside, but used by "mclex()". */

EX int	Line;
/* It is the counter for the number of input line that have been read.
   It is incremented by "lex()", and used when printing error messages.
   It is external because "lex()" may be called to parse several files */

EX int	Chrd;
/* It is the first not yet consumed input char.
   It is initialized outside, but used by "lex()". */

EX double	Vdouble;
/* the value of the double variable just read by "lex()" */

EX char		Buffer[MAXBUFFER];
/* the value of the string variable just read by "lex()" */
# undef EX

int	lex(FILE*);

int E_specexp(char*);

int E_charexp(int);

int E_valerr(char*);

int E_unknown(char*);

void initlex();

    

