
/*************************************************************************

FILE:	main.c

It contains the main procedure, calling the various part of the program.
The declaration of the global variables appears here.

FUNCTIONS CONTAINED:
	main()
*/

#define	MAIN

#include	<stdio.h>
#include	<stdlib.h>
#include	"port.h"
#include	"const.h"
#include	"type.h"
#include	"var.h"
#include	"cspl.h"
#include	"print.h"
#include	"reach.h"
#include	"rdc.h"
#include	"spsor.h"
#include	"lex.h"
#include	"rglang.h"
#include	"tsunif.h"
#include	"tsufsens.h"
#include	"net.h"


int main(argc,argv)
int	argc;
char	*argv[];

/*************************************************************************
The main procedure.
It opens the files and manages the main phases of the analysis.
*/

{
extern	void	compute_holding();
extern	void	adjust_prob();
extern	void	mat_vec_mult();
	int	exitflag;
	char	buf[256];
	FILE	*ff;
	int	method;
	double *temp_prob;
	double prec;
	int iter, meth;
	int i;

	if (argc != 2) {
		fprintf(stderr,"Exactly one parameter required\n");
		exit(ER_CODE);
	}

	model_name = argv[1];

	sprintf(buf,"%s.out",argv[1]);
	if ((Outfile = fopen(buf,"w")) == NULL) {
		fprintf(stderr,"ERROR: cannot open file %s in w(rite) mode\n",buf);
		exit(ER_CODE);
	}

	if (install() != RES_NOERR)
		exit(ER_CODE);
	ac_init();
	Direction = VAL_FROMTO;

	/* Generate the complete reachability graph. The reachability tree is
	stored as a "canonical list" pointed by "Pcan", and as an AVL-tree
	pointed by "Proot".  The reachability graph is obtained by following
	the rows from each marking.  The column info is stored in the field
	"col.mk", that is, it is a pointer to the correspondent marking.
	Absorbing markings have not yet the return to M0 entry.  The vanloops
	are pointed by the a list starting at "vloop".  The first marking and
	the length is stored for each loop , the markings of a loop have
	increasing indexes in absolute value.  */

	exitflag = reach_graph();

	pr_short_rg();
	ac_reach();
	if (IOpt[IOP_MC].val == VAL_CTMC) {
		IOpt[IOP_PR_RGRAPH].val = VAL_YES;
	}
	if (IOpt[IOP_PR_RGRAPH].val == VAL_YES) {
		sprintf(buf,"%s.rg",argv[1]);
		if((ff = fopen(buf,"w")) == NULL) {
			fprintf(stderr,"ERROR: cannot open file %s in w(rite) mode\n",buf);
			exit(ER_CODE);
		}
		pr_long_rg(ff);
		fclose(ff);
		if (exitflag == RES_ERROR)
			exit(ER_CODE);
	}

	/* the analysis method is not transient analysis.  This
	 * initialization is used to prevent transient analysis from being
	 * done if the MC value is set to VAL_DTMC
	 */
	Method_transient = VAL_NO;
	Method_sensitivity = VAL_NO;

	/* Transform the graph into a matrix. The reachability graph is
	transformed into a matrix.  The matrix is stored columnwise (pointed
	by "Qcol", "row.ind" gives the row for each entry). */

	if ( IOpt[IOP_SENSITIVITY].val == VAL_YES  &&
	     IOpt[IOP_MC].val == VAL_CTMC )
	   Method_sensitivity = VAL_YES;

	if (create_matrix() == RES_ERROR)
		exit(ER_CODE);

	if (test_transient() == RES_ERROR)
		exit(ER_CODE);

	compute_holding();
	if (IOpt[IOP_MC].val == VAL_DTMC) {
		Find = -Nvan;
		Nstates = Nvan + Ntottan;
	} else {

		if ( ( IOpt[IOP_METHOD].val == VAL_SSSOR ) ||
		     ( IOpt[IOP_METHOD].val == VAL_GASEI ))
		     Method_transient = VAL_NO;
		else
		     Method_transient = VAL_YES;

        /* The initial probability is calculated only if the initial marking
	 * is vanishing and transient analysis is needed.  If the initial
	 * marking is tangible, its initial probability is 1.
	 */
                if ( Method_transient == VAL_YES && Pcan->index < 0 )
		{
                   /* allocate space for the initial probability vector */
		   Prob_init = (double *)new_mem((unsigned int)
				   (sizeof(double)*(Nvan+Ntottan+1)));
                   Prob_init = &Prob_init[Nvan];
                }

		/* Reduce reach. graph to a CTMC (tang-to-tang) */
		if (rg2ctmc() == RES_ERROR)
			exit(ER_CODE);
		Find = 0;
		Nstates = Ntottan;
	}
	if (IOpt[IOP_PR_MC].val == VAL_YES) {
		sprintf(buf,"%s.mc",argv[1]);
		if ((Mcfile = fopen(buf,"w")) == NULL) {
			fprintf(stderr,"ERROR: cannot open file %s in w(rite) mode\n",buf);
			exit(ER_CODE);
		}
		pr_matrix(Mcfile);
	}

	/* if the evaluation is for steady state */
	if ( Method_transient == VAL_NO ) {

	/* this is to close the .mc file if it is open */
	if (IOpt[IOP_PR_MC].val == VAL_YES)
	   fclose(Mcfile);

	fprintf(stderr,"Solving the Markov chain...\n");
	put_in_vector();/* normalize if DTMC, allocate and initialize D/Cprob */
	Precision = FOpt[FOP_PRECISION].val;
	Iterations = IOpt[IOP_ITERATIONS].val;
	method = IOpt[IOP_METHOD].val;
	if (IOpt[IOP_MC].val == VAL_CTMC) {
		sor(IP,JA,A,Find,Ntottan-1,CProb,&Precision,&Iterations,&method);

		if ( Method_sensitivity == VAL_YES )
		{
		   prec = FOpt[FOP_PRECISION].val;
		   iter = IOpt[IOP_ITERATIONS].val;
		   meth = IOpt[IOP_METHOD].val;
		   temp_prob = (double *) new_mem((unsigned int)
					  (sizeof(double)*Ntottan));

		   mat_vec_mult(CProb,DIP, DJA, DA, Ntottan-1, temp_prob );

		   for ( i = 0; i < Ntottan; i++ )
		   {
			if ( temp_prob[i] != 0.0 )
			   temp_prob[i] = -temp_prob[i];
			printf("tempprob[%d] = %g\n", i, temp_prob[i] );
		   }

		   newsor(IP,JA,A,Find,Ntottan-1,DerCProb,temp_prob,CProb,
			   &prec,&iter,&meth);

		   /* debugging */
		   fprintf(stderr,"Derivatives of the probabilities\n");
		   for ( i = 0; i < Ntottan; i++ )
		      fprintf(stderr,"dprob[%d] = %.12g\n", i, DerCProb[i]);
		   /* debugging */
		}

	}  else {
		sor(IP,JA,A,Find,Ntottan-1,DProb,&Precision,&Iterations,&method);
	}
	if (Precision > FOpt[FOP_PRECISION].val) {
		fprintf(stderr,"ERROR/WARNING: required precision not reached\n");
	}
	fprintf(stderr,"...Markov chain solved\n");
	if (IOpt[IOP_MC].val == VAL_CTMC) {
		release_matrix();
		initlex();
		/* Read the reachability graph info */
		sprintf(buf,"%s.rg",argv[1]);
		if ((ff = fopen(buf,"r")) == NULL) {
			fprintf(stderr,"ERROR: cannot open file %s in r(ead) mode\n",buf);
			exit(ER_CODE);
		}
		fprintf(stderr,"Reading the reachability graph info ...\n");
		if (rg_parse(ff,IOpt[IOP_PR_FULL_MARK].val) == RES_ERROR) {
			fprintf(stderr,"ERROR: reachability graph not read\n");
			exit(ER_CODE);
		}
		fclose(ff);
	} else {
		adjust_prob(DProb,CProb,Holding);
	}
	if (IOpt[IOP_PR_PROB].val == VAL_YES) {
		sprintf(buf,"%s.prb",argv[1]);
		if ((Prbfile = fopen(buf,"w")) == NULL) {
			fprintf(stderr,"ERROR: cannot open file %s in w(rite) mode\n",buf);
			exit(ER_CODE);
		}
		pr_probs(Prbfile,0,Ntottan-1,CProb,method,Precision,Iterations);
		fclose(Prbfile);
	}
	} /* if method transient == val no */

	ac_final();
	if ( Method_transient == VAL_YES )
	{
	   fclose(Mcfile);
	   fclose(Prbfile);
	}

	fprintf(stderr,"End of execution.\n");
	exit(OK_CODE);
} /* main() */

/*
 * time_value()
 *
 * this function is called from ac_final.  It supplies the time at which the
 * transient analysis needs to be carried out.  The functions following this
 * function call in ac_final are evaluated with the probabilities obtained
 * for this time value.
 */

void time_value ( time )
double time;
{
	int method;
	static int first_time = VAL_YES;

	char	buf[256];
	FILE *ff;

	if ( Method_transient == VAL_NO )
		return;

	fprintf(stderr,"Solving the Markov chain...\n");
	if ( first_time == VAL_YES )
	   put_in_vector(); /* allocate and initialize Cprob and Cumprob */

	/* initialize the global time variable.  This is used in cspl.c for
	   computing time averaged values */
	Time = time;

	initialize_probs();

	Precision = FOpt[FOP_PRECISION].val;
	method = IOpt[IOP_METHOD].val;

	if ( IOpt[IOP_PR_MC].val == VAL_YES )
	   pr_time_in_mcfile(Mcfile, time);

	/* call the uniformization routine to analyze the ctmc */
	if ( Method_sensitivity == VAL_NO )
	   uniform(IP,JA,A,Nstates,Nentries,CProb,Cumprob,
		&Precision,time,&method);
	else
	   unif_sens(IP,JA,A,Nstates,Nentries,DIP,DJA,DA,Nderentries,CProb,
	   	Cumprob,DerCProb,DerCumprob,&Precision,time,&method);

	/* debugging
	if ( Method_sensitivity == VAL_YES )
		for ( i = 0; i < Nstates; i++ )
	   		fprintf(stderr,"dprob[%d] = %.12f\n", i, DerCProb[i] );
	debugging */

	fprintf(stderr,"...Markov chain solved\n");
	if (first_time == VAL_YES) {
		first_time = VAL_NO;
		release_matrix();
		initlex();
		sprintf(buf,"%s.rg",model_name);
		if ((ff = fopen(buf,"r")) == NULL) {
			fprintf(stderr,"ERROR: cannot open file %s in r(ead) mode\n",buf);
			exit(ER_CODE);
		}
		fprintf(stderr,"Reading the reachability graph info ...\n");
		if (rg_parse(ff,IOpt[IOP_PR_FULL_MARK].val) == RES_ERROR) {
			fprintf(stderr,"ERROR: reachability graph not read\n");
			exit(ER_CODE);
		}
		fclose(ff);

		if (IOpt[IOP_PR_PROB].val == VAL_YES) {

			sprintf(buf,"%s.prb",model_name);
			if ((Prbfile = fopen(buf,"w")) == NULL) {
				fprintf(stderr,"ERROR: cannot open file %s in w(rite) mode\n",buf);
				exit(ER_CODE);
			}
			pr_probs(Prbfile,0,Ntottan-1,CProb,
			       method,Precision,Iterations);
		} /* if IOpt[IOP_PR_PROB].val */

	}

	if (IOpt[IOP_PR_PROB].val == VAL_YES)
		add_to_prbfile(Prbfile, time, 0 );

	/* print the time value in .mc file */
	print_time ( time );

} /* time_value */

void mat_vec_mult( vect, ip, ja, a, lind, outvec )

double vect[], a[], outvec[];
int ip[], ja[];
int lind;

{
   int index,index2;
   double *d;

   fprintf(stderr,"Debug: mat_vec_mult()\n");

   d = (double *)new_mem((unsigned int)(sizeof (double) * Ntottan));

   for ( index = 0; index <= lind; index++ )
      for ( index2 = ip[index]; index2 < ip[index+1]; index2++ )
	 d[ja[index2]] -= a[index2];

   for ( index = 0; index <= lind; index++ )
   {
      /* diagonal element */
      outvec[index] = vect[index] * d[index];

      for ( index2 = ip[index]; index2 < ip[index+1]; index2++ )
      {
	 outvec[index] += vect[ja[index2]] * a[index2];
	 
      }
   }

   free( d );
}
