/**
FILE:	print.c

It manages the output of data and messages for the whole program.

FUNCTIONS CONTAINED:
	pr_place()
	pr_trans()
	pr_placenames()
	pr_mark()
	pr_row()
	pr_slex()
	pr_stanlex()
	pr_svanlex()
	pr_rtanset()
	pr_rset()
	pr_glex()
	pr_gtanlex()
	pr_gvanlex()
	pr_rgraph()
	pr_reachinfo()
	pr_matrix()
	release_matrix()
	pr_probs()
*/

# include	<stdio.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"
# include	"var.h"
# include	"utility.h"
# include	"rdc.h"


char	*pr_place(buf,pl)
char	*buf;
PL	*pl;

/*************************************************************************
Print the name of the place "pl" into the string "buf".
*/

{
	if (pl->n1 == 1 && pl->n2 == 1)
		sprintf(buf,"%s",pl->name);
	else if (pl->n2 == 1)
		sprintf(buf,"%s.%d",pl->name,pl->i1);
	else
		sprintf(buf,"%s.%d.%d",pl->name,pl->i1,pl->i2);
	return(buf);
}

char	*pr_trans(buf,tr)
char	*buf;
TR	*tr;

/*************************************************************************
Print the name of the transition "tr" into the string "buf".
*/

{
	if (tr->n1 == 1 && tr->n2 == 1)
		sprintf(buf,"%s",tr->name);
	else if (tr->n2 == 1)
		sprintf(buf,"%s.%d",tr->name,tr->i1);
	else
		sprintf(buf,"%s.%d.%d",tr->name,tr->i1,tr->i2);
	return(buf);
}

void pr_placenames(file)
FILE	*file;

/*************************************************************************
Print the name of the places to "file", as an header when PR_FULL_MARK is on.
*/

{
	int	i;
	char	aux[30];

	fprintf(file,"#\t");
	for (i = 0; i < Nplace; i++)
		fprintf(file,"%-7.7s\t",pr_place(aux,&Arrplace[i]));
	fprintf(file,"\n");
}

void pr_mark(pmark,file,use)
ME	*pmark; 	/* pointer to the ME where the marking is stored */
FILE	*file;		/* file where the fprintf will print */
int	use;		/* USE_NAME or USE_NUMBER */

/*************************************************************************
It prints the marking stored in the ME pointed by pmark
*/

{
	char	bufpl[30];
	int	i,count;

	fprintf(file,"%d_%c",pmark->index,pmark->type);

	if (IOpt[IOP_PR_FULL_MARK].val == VAL_YES) {
		for (i = 0; i < Nplace; i++) {
			if (pmark->mk[i] == 0)
				fprintf(file,"\t  :");
			else
				fprintf(file,"\t%3d",pmark->mk[i]);
		}
		fprintf(file,"\n");
	} else {
		for (count = 0, i = 0; i < Nplace; i++) {
			if (pmark->mk[i] == 0)
				continue;
			if (count++ == 9) {
				fprintf(file,"\n");
				count = 1;
			}
			if (use == USE_NUMBER)
				fprintf(file,"\t%3d:%2d",i,pmark->mk[i]);
			else
				fprintf(file,"\t%s:%d",pr_place(bufpl,&Arrplace[i]),pmark->mk[i]);
		}
		fprintf(file,";\n");
	}
} /* pr_mark() */

void pr_row(pmark,file,use)
ME	*pmark;		/* pointer to the ME from where the row of EE starts */
FILE	*file;		/* file where the fprintf must print */
int	use;		/* USE_NAME or USE_NUMBER */

/*************************************************************************
It prints a row of the transition matrix.
The column info is in the format of a pointer to the marking.
It does not print anything if "pmark->prow == NULL" (absorbing marking).
*/

{
	char	buftr[30];
	EE	*auxee;
	int	count;

	if (pmark->prow == NULL)
		return;
	fprintf(file,"%d",pmark->index);
	for (count = 0, auxee = pmark->prow; auxee != NULL; auxee = auxee->row.pt) {
		if (count++ == 2) {
			fprintf(file,"\n");
			count = 1;
		}
		if (use == USE_NUMBER)
			fprintf(file,"\t%4d:%3d:%-.12le",auxee->col.mk->index,auxee->tr,auxee->value);
		else
			fprintf(file,"\t%d:%s:%-.12le",auxee->col.mk->index,pr_trans(buftr,&Arrtrans[auxee->tr]),auxee->value);
	}
	fprintf(file,";\n");
} /* pr_row() */

void pr_slex(root,file)
ME	*root;
FILE	*file;

/*************************************************************************
pr_slex(root,file)
pr_stanlex(root,file)
pr_svanlex(root,file)

Recursively prints the reach. set starting at root, in lexical order.
pr_slex:	all the reach set
pr_stanlex:	the tangible portion
pr_svanlex:	the vanishing portion

root		root of the (sub)tree to print.
file		file where to print the info.
*/

{
	if (root == NULL)
		return;
	pr_slex(root->plr[CMP_LT],file);
	pr_mark(root,file,IOpt[IOP_USENAME].val);
	pr_slex(root->plr[CMP_GT],file);
} /* pr_slex() */


void pr_stanlex(root,file)
ME	*root;
FILE	*file;
{
	if (root == NULL)
		return;
	pr_stanlex(root->plr[CMP_LT],file);
	if (root->index > 0)
		pr_mark(root,file,IOpt[IOP_USENAME].val);
	pr_stanlex(root->plr[CMP_GT],file);
} /* pr_stanlex() */


void pr_svanlex(root,file)
ME	*root;
FILE	*file;
{
	if (root == NULL)
		return;
	pr_svanlex(root->plr[CMP_LT],file);
	if (root->index < 0)
		pr_mark(root,file,IOpt[IOP_USENAME].val);
	pr_svanlex(root->plr[CMP_GT],file);
} /* pr_svanlex() */

void pr_rtanset(rgfile)
FILE	*rgfile;	/* file where the info is printed */

/*************************************************************************
Prints the tangible reachability set.
*/

{
	ME	*auxme;

	if (IOpt[IOP_PR_FULL_MARK].val == VAL_YES)
		pr_placenames(rgfile);
	switch (IOpt[IOP_PR_MARK_ORDER].val) {
	case VAL_CANONIC:
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->index > 0)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		break;
	case VAL_LEXICAL:
		pr_stanlex(Proot,rgfile);
		break;
	case VAL_MATRIX:
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_TAN)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_ABS)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		break;
	}
}

void pr_rset(rgfile)
FILE	*rgfile;	/* file where the info is printed */

/*************************************************************************
Prints the information obtained during the generation of the reachability set.
*/

{
	ME	*auxme;

	if (IOpt[IOP_PR_FULL_MARK].val == VAL_YES)
		pr_placenames(rgfile);
	switch (IOpt[IOP_PR_MARK_ORDER].val) {
	case VAL_CANONIC:
		if (IOpt[IOP_PR_MERG_MARK].val == VAL_YES) {
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		} else {
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				if (auxme->index < 0)
					pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				if (auxme->index > 0)
					pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		}
		break;
	case VAL_LEXICAL:
		if (IOpt[IOP_PR_MERG_MARK].val == VAL_YES) {
			pr_slex(Proot,rgfile);
		} else {
			pr_stanlex(Proot,rgfile);
			pr_svanlex(Proot,rgfile);
		}
		break;
	case VAL_MATRIX:
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_VAN || auxme->type == MAR_VLO)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_TAN)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_ABS)
				pr_mark(auxme,rgfile,IOpt[IOP_USENAME].val);
		break;
	}
}

void pr_glex(root,file)
ME	*root;
FILE	*file;

/*************************************************************************
pr_glex(root,file)
pr_gtanlex(root,file)
pr_gvanlex(root,file)

Recursively prints the reach. set starting at root, in lexical order.
pr_glex:		all the reach set
pr_gtanlex:	the tangible portion
pr_gvanlex:	the vanishing portion

root		root of the (sub)tree to print.
file		file where to print the info.
*/

{
	if (root == NULL)
		return;
	pr_glex(root->plr[CMP_LT],file);
	pr_row(root,file,IOpt[IOP_USENAME].val);
	pr_glex(root->plr[CMP_GT],file);
} /* pr_glex() */


void pr_gtanlex(root,file)
ME	*root;
FILE	*file;
{
	if (root == NULL)
		return;
	pr_gtanlex(root->plr[CMP_LT],file);
	if (root->index > 0)
		pr_row(root,file,IOpt[IOP_USENAME].val);
	pr_gtanlex(root->plr[CMP_GT],file);
} /* pr_gtanlex() */


void pr_gvanlex(root,file)
ME	*root;
FILE	*file;
{
	if (root == NULL)
		return;
	pr_gvanlex(root->plr[CMP_LT],file);
	if (root->index < 0)
		pr_row(root,file,IOpt[IOP_USENAME].val);
	pr_gvanlex(root->plr[CMP_GT],file);
} /* pr_gvanlex() */

void pr_rgraph(rgfile)
FILE	*rgfile;	/* file where the info is printed */

/*************************************************************************
It prints the reachability graph by rows.
*/

{
	ME	*auxme;

	switch (IOpt[IOP_PR_MARK_ORDER].val) {
	case VAL_CANONIC:
		if (IOpt[IOP_PR_MERG_MARK].val == VAL_YES) {
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				pr_row(auxme,rgfile,IOpt[IOP_USENAME].val);
		} else {
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				if (auxme->index < 0)
					pr_row(auxme,rgfile,IOpt[IOP_USENAME].val);
			for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
				if (auxme->index > 0)
					pr_row(auxme,rgfile,IOpt[IOP_USENAME].val);
		}
		break;
	case VAL_LEXICAL:
		if (IOpt[IOP_PR_MERG_MARK].val == VAL_YES) {
			pr_glex(Proot,rgfile);
		} else {
			pr_gtanlex(Proot,rgfile);
			pr_gvanlex(Proot,rgfile);
		}
		break;
	case VAL_MATRIX:
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_VAN || auxme->type == MAR_VLO)
				pr_row(auxme,rgfile,IOpt[IOP_USENAME].val);
		for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan)
			if (auxme->type == MAR_TAN)
				pr_row(auxme,rgfile,IOpt[IOP_USENAME].val);
		break;
	}
} /* pr_rgraph() */

void pr_short_rg()

/*************************************************************************
Prints on "stderr" the overall information obtained during the generation
of the reachability graph.
*/

{
	LL	*auxll;
	int	i;

	fprintf(stderr,"The reachability graph contains:\n");
	fprintf(stderr,"%9d tangible markings",Ntottan);
	if (Nabs > 0)
		fprintf(stderr," (%d absorbing)",Nabs);
	fprintf(stderr,"\n%9d vanishing markings",Nvan);
	fprintf(stderr,"\n");
	for (i = 0, auxll = Vloop; auxll != NULL; auxll = auxll->pnext)
		i++;
	if (i > 0)
		fprintf(stderr,"%9d vanloops\n",i);
	fprintf(stderr,"%9d arcs\n",Narcs);
}

void pr_long_rg(rgfile)
FILE	*rgfile;

/*************************************************************************
Prints the information obtained during the generation of the reachability graph.
*/

{
	char	buf[30];
	LL	*auxll;
	int	i;

	/* printf the conversion table, from number to name */
	fprintf(rgfile,"_nplace = %d;\n",Nplace);
	fprintf(rgfile,"_ntrans = %d;\n",Ntrans);
	fprintf(rgfile,"_places =\n");
	for (i = 0; i < Nplace; i++)
		fprintf(rgfile,"\t%d: %s;\n",i,pr_place(buf,&Arrplace[i]));
	fprintf(rgfile,"_transitions =\n");
	for (i = 0; i < Ntrans; i++)
		fprintf(rgfile,"\t%d: %s;\n",i,pr_trans(buf,&Arrtrans[i]));

	/* print the reachability graph info */
	fprintf(rgfile,"_ntanmark = %d;\n",Ntan);
	fprintf(rgfile,"_nabsmark = %d;\n",Nabs);
	fprintf(rgfile,"_nvanmark = %d;\n",Nvan);
	for (i = 0, auxll = Vloop; auxll != NULL; auxll = auxll->pnext)
		i++;
	fprintf(rgfile,"_nvanloop = %d;\n",i);
	fprintf(rgfile,"_nentries = %d;\n",Narcs);

	switch (IOpt[IOP_PR_RSET].val) {
	case VAL_YES:
		fprintf(rgfile,"_reachset =\n");
		pr_rset(rgfile);
		break;
	case VAL_NO:
		break;
	case VAL_TANGIBLE:
		fprintf(rgfile,"_reachset =\n");
		pr_rtanset(rgfile);
		break;
	}

	switch (IOpt[IOP_PR_RGRAPH].val) {
	case VAL_YES:
		fprintf(rgfile,"_reachgraph =\n");
		pr_rgraph(rgfile);
		break;
	case VAL_NO:
		break;
	}
}

void pr_matrix(mcfile)
FILE	*mcfile;	/* file where the fprintf will print */

/*************************************************************************
It prints the matrix in rowwise or columnwise order.
At the end the matrix is in the requested order for printing.
*/

{
extern	void	row2col();
extern	void	col2row();
	int	col,row,count;
	EE	*auxee;
	int	index;

	/* change order of entries, if so desired */
	if (IOpt[IOP_PR_MC_ORDER].val == VAL_TOFROM && Direction == VAL_FROMTO) {
		row2col(-Nvan,Ntottan-1);
		Direction = VAL_TOFROM;
	} else if (IOpt[IOP_PR_MC_ORDER].val == VAL_FROMTO && Direction == VAL_TOFROM) {
		col2row(-Nvan,Ntottan-1);
		Direction = VAL_FROMTO;
	}
	if ( Method_sensitivity == VAL_YES )
	{
		if (IOpt[IOP_PR_MC_ORDER].val == VAL_TOFROM &&
			DerDirection == VAL_FROMTO)
		{
			derrow2dercol(-Nvan,Ntottan-1);
			DerDirection = VAL_TOFROM;
		}
		else if (IOpt[IOP_PR_MC_ORDER].val == VAL_FROMTO &&
			DerDirection == VAL_TOFROM)
		{
			dercol2derrow(-Nvan,Ntottan-1);
			DerDirection = VAL_FROMTO;
		}
	}

	fprintf(mcfile,"_firstindex = %d;\n",Find);
	fprintf(mcfile,"_nstates = %d;\n",Nstates);
	fprintf(mcfile,"_nentries = %d;\n",Nentries);
	if (IOpt[IOP_PR_MC_ORDER].val == VAL_TOFROM)
		fprintf(mcfile,"_order = _TOFROM;\n");
	else
		fprintf(mcfile,"_order = _FROMTO;\n");
	if (Nentries != 0)
		fprintf(mcfile,"_matrix =\n");
	if (IOpt[IOP_PR_MC_ORDER].val == VAL_FROMTO) {
		for (row = -Nvan; row < Ntottan; row++) {
			if (Qrow[row] == NULL)
				continue;
			fprintf(mcfile,"%4d",row);
			for (count = 0, auxee = Qrow[row]; auxee != NULL; auxee = auxee->row.pt) {
				if (count++ == 3) {
					fprintf(mcfile,"\n");
					count = 1;
				}
				fprintf(mcfile,"\t%4d:%-.12le",auxee->col.ind,
				(row >= 0 && IOpt[IOP_MC].val == VAL_DTMC)
				? auxee->value * Holding[row]
				: auxee->value);
			}
			fprintf(mcfile,";\n");
		}
	} else {
		for (col = -Nvan; col < Ntottan; col++) {
			if (Qcol[col] == NULL)
				continue;
			fprintf(mcfile,"%4d",col);
			for (count = 0, auxee = Qcol[col]; auxee != NULL; auxee = auxee->col.pt) {
				if (count++ == 3) {
					fprintf(mcfile,"\n");
					count = 1;
				}
				fprintf(mcfile,"\t%4d:%-.12le",auxee->row.ind,
				(auxee->row.ind >= 0 && IOpt[IOP_MC].val == VAL_DTMC)
				? auxee->value * Holding[auxee->row.ind]
				: auxee->value);
			}
			fprintf(mcfile,";\n");
		}
	}
	if ( Method_sensitivity == VAL_YES )
	{
	if (Nderentries != 0)
		fprintf(mcfile,"_dermatrix =\n");
	if (IOpt[IOP_PR_MC_ORDER].val == VAL_FROMTO) {
		for (row = -Nvan; row < Ntottan; row++) {
			if (DerQrow[row] == NULL)
				continue;
			fprintf(mcfile,"%4d",row);
			for (count = 0, auxee = DerQrow[row]; auxee != NULL; auxee = auxee->row.pt) {
				if (count++ == 3) {
					fprintf(mcfile,"\n");
					count = 1;
				}
				fprintf(mcfile,"\t%4d:%-.12le",auxee->col.ind, auxee->d_value);
			}
			fprintf(mcfile,";\n");
		}
	} else {
		for (col = -Nvan; col < Ntottan; col++) {
			if (DerQcol[col] == NULL)
				continue;
			fprintf(mcfile,"%4d",col);
			for (count = 0, auxee = DerQcol[col]; auxee != NULL; auxee = auxee->col.pt) {
				if (count++ == 3) {
					fprintf(mcfile,"\n");
					count = 1;
				}
				fprintf(mcfile,"\t%4d:%-.12le",auxee->row.ind, auxee->d_value);
			}
			fprintf(mcfile,";\n");
		}
	}
	} /* if method sensitivity */
	switch (IOpt[IOP_METHOD].val) {
	case VAL_GASEI: fprintf(mcfile,"_method = _GASEI;\n"); break;
	case VAL_SSSOR: fprintf(mcfile,"_method = _SSSOR;\n"); break;
	case VAL_TSUNIF: fprintf(mcfile,"_method = _TSUNIF;\n"); break;
	}
	fprintf(mcfile,"_precision = %lg;\n",FOpt[FOP_PRECISION].val);

	if ( Method_transient == VAL_NO )
	{
	   fprintf(mcfile,"_iterations = %d;\n",IOpt[IOP_ITERATIONS].val);
	   fprintf(mcfile,"_solve = _ALL;\n");
	}
	else if ( Method_transient == VAL_YES && Pcan->index < 0 )
	{
   	   fprintf(mcfile,"_initstate = \n");

   	   count = 1;
   	   for( index = 0; index < Ntottan; index++ )
   	   {
      	      if ( Prob_init[index] > 0.0 )
      	      {
	         /* only two values printed on a single line */
                 if ( count == 3 )
                 {
	            fprintf(mcfile,"\n");
	            count = 1;
                 }
	         count++;
	         fprintf(mcfile,"\t%d:%-.12g", index, Prob_init[index]);
      	      }
   	   } /* for index */
   	   fprintf(mcfile,";\n");

	} /* else if method transient */

} /* pr_matrix() */

void release_matrix()

/*************************************************************************
It releases the matrix entries.
*/

{
	int	col,row;

	if (Direction == VAL_TOFROM) {
		for (col = -Nvan; col < Ntottan; col++) {
			if (Qcol[col] == NULL)
				continue;
			rel_EE(Qcol[col],VAL_TOFROM);
			Qcol[col] = NULL;
		}
	} else {
		for (row = -Nvan; row < Ntottan; row++) {
			if (Qrow[row] == NULL)
				continue;
			rel_EE(Qrow[row],VAL_FROMTO);
			Qrow[row] = NULL;
		}
	}
} /* release_matrix() */

void pr_probs(prbfile,find,lind,prob,method,precision,iterations)
FILE	*prbfile;	/* the file for the output */
int	find;		/* the index of the first state of the MC */
int	lind;		/* the index of the last state of the MC */
double	prob[];		/* the vector of probabilities */
int	method;		/* the type of solution */
double	precision;	/* the precision value */
int	iterations;	/* the iterations value */

/*************************************************************************
It prints the output info to "prbfile".
*/

{
	int	i,j,nstates;

	nstates = lind - find + 1;
	fprintf(prbfile,"_firstindex = %d;\n",find);
	fprintf(prbfile,"_nstates = %d;\n",nstates);
	switch (IOpt[IOP_METHOD].val) {
	case VAL_GASEI: fprintf(prbfile,"_method = _GASEI;\n"); break;
	case VAL_SSSOR: fprintf(prbfile,"_method = _SSSOR;\n"); break;
	case VAL_TSUNIF: fprintf(prbfile,"_method = _TSUNIF;\n"); break;
	}

	fprintf(prbfile,"_precision = %.12le;\n",precision);
	if ( Method_transient == VAL_NO )
	{
	   fprintf(prbfile,"_iterations = %d;\n",iterations);
	   fprintf(prbfile,"_probabilities =");
	   for (j = 0, i = find; i <= lind; i++)
		   if (prob[i] > 0.0)
			   fprintf(prbfile,"%c%d:%.12le",
			   ((j++)%3)?'\t':'\n',i,prob[i]);
	   if ( Method_sensitivity == VAL_YES )
	   {
	   	fprintf(prbfile,";\n_derprobabilities =");
	   	for (j = 0, i = find; i <= lind; i++)
		   	if (DerCProb[i] != 0.0)
			   	fprintf(prbfile,"%c%d:%.12le",
			   	((j++)%3)?'\t':'\n',i,DerCProb[i]);
	   }
	   fprintf(prbfile,";\n_end = ;\n");
	}

} /* end of pr_probs() */

/********************************************************
 *  This is a new function added to put the time points where
 *  the Markov chain should be solved, in the .mc file
 */
void pr_time_in_mcfile( mcfile, time_point )
FILE *mcfile;
double time_point;

{
   if ( Method_transient == VAL_YES )
   {
      fprintf(mcfile,"_time = %.12f;\n", time_point);
      fprintf(mcfile,"_solve = _ALL;\n");
   }
}

void add_to_prbfile( prbfile, time, find )
FILE *prbfile;
double time;
int find;

{
	int i,j;

	fprintf(prbfile,"_time = %.12e;\n",time);
	fprintf(prbfile,"_probabilities =");
	for (j = 0, i = find; i < Nstates; i++)
		if (CProb[i] > 0.0)
			fprintf(prbfile,"%c%d:%.12e",((j++)%3)?'\t':'\n',i,CProb[i]);
	fprintf(prbfile,";\n");
	fprintf(prbfile,"_cumprobabilities =");
	for (j = 0, i = find; i < Nstates; i++)
		if (Cumprob[i] > 0.0)
			fprintf(prbfile,"%c%d:%.12e",
			    ((j++)%3)?'\t':'\n',i,Cumprob[i]);
	if ( Method_sensitivity == VAL_YES )
	{
		fprintf(prbfile,";\n");
		fprintf(prbfile,"_derprobabilities =");
		for (j = 0, i = find; i < Nstates; i++)
			if (DerCProb[i] != 0.0)
				fprintf(prbfile,"%c%d:%.12e",
			    		((j++)%3)?'\t':'\n',i,DerCProb[i]);
		fprintf(prbfile,";\n");
		fprintf(prbfile,"_dercumprobabilities =");
		for (j = 0, i = find; i < Nstates; i++)
			if (DerCumprob[i] != 0.0)
				fprintf(prbfile,"%c%d:%.12e",
			    		((j++)%3)?'\t':'\n',i,DerCumprob[i]);
	}
	fprintf(prbfile,";\n_end = ;\n");
}
