
/*************************************************************************

FILE:	rdc.c

It performs operations on the reach. graph after its generartion, then
it reduces the reach. graph to a tangible to tangible matrix.

FUNCTIONS CONTAINED:
	compute_holding()
	test_transient()
	gaussian()
	backsubst()
	checkQtt()
	rg2ctmc()
	row2col()
	col2row()
	adjust_prob()
	put_in_vector()

*/

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"
# include	"var.h"
# include	"utility.h"

void	compute_holding()

/*************************************************************************
Computes the holding times for each (tangible) marking.
*/

{
	int	i,col;
	EE	*auxee;

	if (Direction != VAL_TOFROM)
		exit(23886);
	Holding = (double *) new_mem((unsigned int)(sizeof(double)*Ntan));
	for (i = 0; i < Ntan; i++)
		Holding[i] = 0.0;
	for (col = -Nvan; col < Ntottan; col++)
		for (auxee = Qcol[col]; auxee != NULL; auxee = auxee->col.pt)
			if (auxee->row.ind >= 0)
				Holding[auxee->row.ind] += auxee->value;
	for (i = 0; i < Ntan; i++)
		Holding[i] = 1.0 / Holding[i];
} /* compute_holding() */

int	test_transient()

/*************************************************************************
Test whether the initial marking is transient, and takes the appropriate
actions, according to user option.
RETURNS:
RES_ERROR or RES_NOERR.
*/

{
	int	*list;		/* FIFO list of markings to be considered */
	int	length;		/* dimension of the array to be allocated */
	char	*reach;		/* reach[i] == 1 if marking i reaches m0 */
	int	numreach;	/* # of markings reaching m0 */
	int	head,tail;
	EE	*auxee;

	length = Ntottan + Nvan;
	list = (int *) new_mem((unsigned int)(sizeof(int) * length));
	reach = (char *) new_mem((unsigned int)(sizeof(char) * length));
	for (head = 0; head < length; head++)
		reach[head] = VAL_NO;
	reach = &reach[Nvan];

	/* initialize list */
	numreach = 1;
	head = 0;
	tail = 0;
	list[head] = Pcan->index;
	reach[Pcan->index] = VAL_YES;
	while (head <= tail && numreach < length) {
		for (auxee = Qcol[list[head++]]; auxee != NULL; auxee = auxee->col.pt)
			if (reach[auxee->row.ind] == VAL_NO) {
				reach[auxee->row.ind] = VAL_YES;
				list[++tail] = auxee->row.ind;
				numreach++;
			}
	}
	if (numreach < length) {
		if (IOpt[IOP_OK_TRANS_M0].val == VAL_NO) {
			fprintf(stderr,"ERROR: transient initial marking. %d markings reach it.\n",--numreach);
			return(RES_ERROR);
		} else
			fprintf(stderr,"ERROR/WARNING: transient initial marking. %d markings reach it.\n",--numreach);
    }
	free((char *) list);
	free((char *) &reach[-Nvan]);
	return(RES_NOERR);
} /* test_transient() */


void gaussian(vector,first,last)
EE	**vector;	/* pointer to the rows of the matrix */
int	first;		/* index in vector of the first row of the matrix */
int	last;		/* index in vector of the last row of the matrix */

/*************************************************************************
Gaussian Elimination is used, without pivoting.
The initial situation is a matrix whose rows are pointed by vector,
and ordered, the relevant rows have indexes between first and last.
The matrix contain the square matrix A the subsequent right hand sides B.
At the end, if A = LU, vector points to the rows of [M U | M L**-1 B],
where M is a diagonal matrix needed to have the diagonal entries of [M U]
equal to -1 (M[i][i] = 1/U[i][i]).
The diagonal element of A, and of [M U], are -1, are not stored.
In the middle of the computation (i.e. at step i) certain diagonal
elements (A[j][j], j>i) can be esplicitly stored.
In this case A[j][j]+1 is the value actually stored, that is, the -1
value is added only at step j.
*/

{
	int	j;
	double	mult;
	EE	*auxee;
	EE	*pi,*pj,*oldpj;
#ifdef VERBOSE
	int	Noper = 0;	/* counts the operations */
	int	Narit = 0;	/* counts the arithmetic operations */
#endif

	for (j = first; j <= last; j++) {
#ifdef VERBOSE
	Noper++;
#endif
		while (vector[j]->col.ind < j) {
#ifdef VERBOSE
			Noper++;
#endif
			/* eliminate entry (j,i): ROWj += ROWi * mult */
			mult = vector[j]->value;
			oldpj = vector[j];
			pj = oldpj->row.pt;
			for (pi = vector[oldpj->col.ind]; pi != NULL;
			pi = pi->row.pt) {
				 while (pj != NULL && pj->col.ind < pi->col.ind) {
#ifdef VERBOSE
					Noper++;
#endif
					pj = (oldpj = pj)->row.pt;
				}
				 if (pj == NULL) {
					for (; pi != NULL; pi = pi->row.pt) {
#ifdef VERBOSE
					Noper++;
					Narit++;
#endif
						oldpj->row.pt = get_EE();
						oldpj = oldpj->row.pt;
						oldpj->row.pt = NULL;
						oldpj->col.ind = pi->col.ind;
						oldpj->value = pi->value * mult;
					}
					break;
				 }
#ifdef VERBOSE
				Noper++;
				Narit++;
#endif
				 if (pj->col.ind == pi->col.ind) {
					/* entry already present, just change the value */
					pj->value += pi->value * mult;
					oldpj = pj;
					pj = oldpj->row.pt;
				} else { /*pj->col.ind > pi->col.ind*/
					/* a new entry is needed */
					oldpj->row.pt = get_EE();
					oldpj = oldpj->row.pt;
					oldpj->row.pt = pj;
					oldpj->col.ind = pi->col.ind;
					oldpj->value = pi->value * mult;
				}
			} /* for pi */
			/* release entry (j,i) */
			auxee = vector[j];
			vector[j] = vector[j]->row.pt;
			auxee->row.pt = NULL;
			rel_EE(auxee,VAL_FROMTO);
		} /* end of while */
		/* normalize row j, if necessary */
		if (vector[j]->col.ind == j) {
			/* the diagonal was changed while eliminating */
			/* release the diagonal element, and normalize */
			mult = -1/(vector[j]->value - 1);
			auxee = vector[j];
			vector[j] = vector[j]->row.pt;
			auxee->row.pt = NULL;
			rel_EE(auxee,VAL_FROMTO);
			for (auxee = vector[j]; auxee != NULL; auxee = auxee->row.pt) {
				auxee->value *= mult;
#ifdef VERBOSE
				Noper++;
				Narit++;
#endif
			}
		}
	} /* for j */
#ifdef	VERBOSE
	fprintf(stderr,"++++++\t\t\tOperations in procedure 'gaussian': %d (+ %d)\n",Narit,Noper);
#endif
} /* gaussian() */

void backsubst()

/*************************************************************************
Given Pvv (by rows, upper triangular), Pvt (by columns, rectangular),
Qtv (by rows, rectangular), Qtt (by columns, square), it computes
			Q := Qtt + Qtv * Pvv**-1 Ptv.
It performs a back substitution algorithm on the matrix Pvv, using Pvt
as the rigth hand sides.
The diagonal values of Pvv are assumed to be -1, and not stored.
At the end multiple entries and diagonal entries may be in Q, even if Qtt
was free of them.
Entries of Qtv, Pvv, and Pvt are released.
Pvv is in Qrow (from -Nvan to -1);
Pvt is in Qcol (from 1 to Ntottan, following the columns of Qtt);
Qtv is in Qrow (from 1 to Ntottan);
Qtt is in Qcol (from 1 to Ntottan, preceding the columns of Ptv);
*/

{
	int	i,j;
	EE	*X;	/* the solution of Pvv * X = Pvt[i] */
	//EE	*DerX;	/* the solution of Pvv * DerX = DerPvt[i] */
	EE	*Pvt;	/* for each i, represents the i-th column Pvt[i] */
	double	innerprod;
	EE	*auxee1,*auxee2;
        EE      *auxee3;
	int     pcan_index;
	int     flag_init = VAL_NO;

#ifdef VERBOSE
	int	Noper = 0;	/* counts the operations */
	int	Narit = 0;	/* counts the arithmetic operations */
#endif

        /* index of the initial marking */
	pcan_index = Pcan->index;

        /* if the method is transient and initial marking is vanishing
	 * then set the flag
	 */

        if ((Method_transient == VAL_YES ) &&
            (pcan_index < 0 ))
	{
	   flag_init = VAL_YES;
	}


	for (i = 0; i < Ntottan; i++) {
#ifdef VERBOSE
		Noper++;
#endif
		/* find the part of Pvt in column i */
		if (Qcol[i] == NULL)
			continue;
		if (Qcol[i]->row.ind < 0) {	/* all V->T column */
			Pvt = Qcol[i];
			Qcol[i] = NULL;
		} else {
			auxee1 = Qcol[i];
			Pvt = auxee1->col.pt;
			while (Pvt != NULL && Pvt->row.ind >= 0) {
				auxee1 = Pvt;
				Pvt = auxee1->col.pt;
			}
			if (Pvt == NULL)	/* no V->T entry */
				continue;
			auxee1->col.pt = NULL;
		}
		/* solve with column i of Pvt as rigth hand side */
		/* the computation of X(-1) through X(Pvt[i]->col.ind)
		is skipped, since these X's are 0 */
		X = Pvt;
		Pvt = Pvt->col.pt;
		X->col.pt = NULL;
		X->value = - X->value;
		for (j = X->row.ind; --j >= -Nvan;) {
#ifdef VERBOSE
			Noper++;
#endif
			innerprod = 0;
			auxee1 = Qrow[j];
			auxee2 = X;
			while (auxee1 != NULL && auxee2 != NULL) {
#ifdef VERBOSE
				Noper++;
#endif
				if (auxee1->col.ind == auxee2->row.ind) {
					innerprod += auxee1->value * auxee2->value;
#ifdef VERBOSE
					Narit++;
#endif
					auxee1 = auxee1->row.pt;
					auxee2 = auxee2->col.pt;
				} else if (auxee1->col.ind < auxee2->row.ind)
					auxee1 = auxee1->row.pt;
				else
					auxee2 = auxee2->col.pt;
			}
			if (Pvt != NULL && Pvt->row.ind == j) {
				/* rhs value is != 0 for row j */
				auxee1 = Pvt;
				Pvt = Pvt->col.pt;
				auxee1->value = innerprod - auxee1->value;
#ifdef VERBOSE
				Narit++;
#endif
				auxee1->col.pt = X;
				X = auxee1;
			} else if (innerprod != 0) {
				auxee1 = get_EE();
				auxee1->row.ind = j;
				auxee1->value = innerprod;
				auxee1->col.pt = X;
				X = auxee1;
			}
		} /* while j */
/* Now -X[j] represents the probability of exiting in tangible marking i,
if vanishing marking j is entered. Now Qtv * -X, must be added to Qtt */
		for (j = 0; j < Ntottan; j++) {
#ifdef VERBOSE
			Noper++;
#endif
			innerprod = 0;
			auxee1 = Qrow[j];
			auxee2 = X;
			while (auxee1 != NULL && auxee2 != NULL) {
#ifdef VERBOSE
				Noper++;
#endif
				if (auxee1->col.ind == auxee2->row.ind) {
					innerprod -= auxee1->value * auxee2->value;
#ifdef VERBOSE
					Narit++;
#endif
					auxee1 = auxee1->row.pt;
					auxee2 = auxee2->col.pt;
				} else if (auxee1->col.ind < auxee2->row.ind)
					auxee1 = auxee1->row.pt;
				else
					auxee2 = auxee2->col.pt;
			}
			if (innerprod != 0) {
				auxee1 = get_EE();
				auxee1->row.ind = j;
				auxee1->value = innerprod;
				auxee1->col.pt = Qcol[i];
				auxee1->tr = MCT_PTH;
				Qcol[i] = auxee1;
			}
		} /* for j */

                /* if initial probability distribution is required */
		if ( flag_init == VAL_YES )
                {
		   /* load the initial probability d_values */
		   Prob_init[i] = 0.0;

		   for ( auxee3 = X; auxee3 != NULL; auxee3 = auxee3->col.pt )
		   {
		      if ( auxee3->row.ind == pcan_index )
			 Prob_init[i] = -auxee3->value;
                    }
                 }

		if ( Method_sensitivity == VAL_YES )
		{
		   /* DerQ = DerQtt + DerQtv * X + Qtv * DerX */

                   /* Now DerQtv * -X, must be added to DerQtt */
		   for (j = 0; j < Ntottan; j++)
		   {
		      innerprod = 0;
		      auxee1 = DerQrow[j];
		      auxee2 = X;
		      while (auxee1 != NULL && auxee2 != NULL)
		      {
		         if (auxee1->col.ind == auxee2->row.ind)
			 {
		      	    innerprod -= auxee1->d_value * auxee2->value;
		      	    auxee1 = auxee1->row.pt;
		      	    auxee2 = auxee2->col.pt;
		      	 }
			 else if (auxee1->col.ind < auxee2->row.ind)
		      	         auxee1 = auxee1->row.pt;
		      	      else
		      	         auxee2 = auxee2->col.pt;
		      } /* while */

		      if (innerprod != 0) {
		      	auxee1 = get_EE();
		      	auxee1->row.ind = j;
		      	auxee1->d_value = innerprod;
		      	auxee1->col.pt = DerQcol[i];
		      	auxee1->tr = MCT_PTH;
		      	DerQcol[i] = auxee1;
		      }
	           } /* for j */

		   /* [I-Pvv] * DerX = DerPvt + DerPvv * X 
		    *
		    * First compute DerPvv * X 
		    * Then add DerPvt and DerPvv * X 
		    * This product might result in additional entries
		    * in DerPv
		    * Then Solve the above equation using Gaussian
		    * elimination just as X was computed.
		    */

                   /* Now DerPvv * -X, must be added to DerPvt */
		   for (j = -Nvan; j < 0; j++)
		   {
		      innerprod = 0;
		      auxee1 = DerQrow[j];
		      auxee2 = X;
		      while (auxee1 != NULL && auxee2 != NULL)
		      {
		         if (auxee1->col.ind == auxee2->row.ind)
			 {
		      	    innerprod -= auxee1->d_value * auxee2->value;
		      	    auxee1 = auxee1->row.pt;
		      	    auxee2 = auxee2->col.pt;
		      	 }
			 else if (auxee1->col.ind < auxee2->row.ind)
		      	         auxee1 = auxee1->row.pt;
		      	      else
		      	         auxee2 = auxee2->col.pt;
		      } /* while */

		      if (innerprod != 0) {
			 auxee1 = DerQcol[i];
			 auxee2 = NULL;
			 while ( auxee1 != NULL &&
				 auxee1->row.ind >= j )
			 {
			    auxee2 = auxee1;
			    auxee1 = auxee1->col.pt;
			 }
			 /* At this point the entry auxee2 will
			  * either be pointing to the actual DerPvt
			  * entry required or an entry with an
			  * index greater than the DerPvt entry required
			  * If the entry is present, just increment the
			  * value, if not create a new entry
			  */
			 /* row of DerPvt is empty */
			 if ( auxee2 == NULL )
			 {
		            auxee1 = get_EE();
		            auxee1->row.ind = j;
		            auxee1->d_value = innerprod;
		            auxee1->col.pt = DerQcol[i];
		            auxee1->tr = MCT_PTH;
		            DerQcol[i] = auxee1;
			 }
			 /* if the entry is present in the column */
			 else if ( auxee2->row.ind == j )
			 {
			    auxee2->d_value += innerprod;
			 }
			 /* entry not present */
			 else
			 {
		            auxee3 = get_EE();
		            auxee3->row.ind = j;
		            auxee3->d_value = innerprod;
		            auxee3->tr = MCT_PTH;
			    auxee3->col.pt = auxee1;
			    auxee2->col.pt = auxee3;
			 }
		      } /* if innerprod != 0 */
	           } /* for j DerPvv * -X */

		   if (X != NULL)	/* release the entries of X */
			   rel_EE(X,VAL_TOFROM);

		   /* Now solve for DerX */
		   /* find the part of Pvt in column i */
		   if (DerQcol[i] == NULL)
			   continue;
		   if (DerQcol[i]->row.ind < 0) {	/* all V->T column */
			   Pvt = DerQcol[i];
			   DerQcol[i] = NULL;
		   } else {
			   auxee1 = DerQcol[i];
			   Pvt = auxee1->col.pt;
			   while (Pvt != NULL && Pvt->row.ind >= 0) {
				   auxee1 = Pvt;
				   Pvt = auxee1->col.pt;
			   }
			   if (Pvt == NULL)	/* no V->T entry */
				   continue;
			   auxee1->col.pt = NULL;
		   }
		   /* solve with column i of Pvt as rigth hand side */
		   /* the computation of X(-1) through X(Pvt[i]->col.ind)
		   is skipped, since these X's are 0 */
		   X = Pvt;
		   Pvt = Pvt->col.pt;
		   X->col.pt = NULL;
		   X->d_value = - X->d_value;
		   for (j = X->row.ind; --j >= -Nvan;) {
			   innerprod = 0;
			   auxee1 = Qrow[j];
			   auxee2 = X;
			   while (auxee1 != NULL && auxee2 != NULL) {
				   if (auxee1->col.ind == auxee2->row.ind) {
					   innerprod += auxee1->d_value * auxee2->d_value;
					   auxee1 = auxee1->row.pt;
					   auxee2 = auxee2->col.pt;
				   } else if (auxee1->col.ind < auxee2->row.ind)
					   auxee1 = auxee1->row.pt;
				   else
					   auxee2 = auxee2->col.pt;
			   }
			   if (Pvt != NULL && Pvt->row.ind == j) {
				   /* rhs d_value is != 0 for row j */
				   auxee1 = Pvt;
				   Pvt = Pvt->col.pt;
				   auxee1->d_value = innerprod - auxee1->d_value;
				   auxee1->col.pt = X;
				   X = auxee1;
			   } else if (innerprod != 0) {
				   auxee1 = get_EE();
				   auxee1->row.ind = j;
				   auxee1->d_value = innerprod;
				   auxee1->col.pt = X;
				   X = auxee1;
			   }
		   } /* for j solving for DerX */

                   /* Now Qtv * -DerX, must be added to DerQtt */
		   for (j = 0; j < Ntottan; j++)
		   {
		      innerprod = 0;
		      auxee1 = Qrow[j];
		      auxee2 = X;
		      while (auxee1 != NULL && auxee2 != NULL)
		      {
		         if (auxee1->col.ind == auxee2->row.ind)
			 {
		      	    innerprod -= auxee1->value * auxee2->d_value;
		      	    auxee1 = auxee1->row.pt;
		      	    auxee2 = auxee2->col.pt;
		      	 }
			 else if (auxee1->col.ind < auxee2->row.ind)
		      	         auxee1 = auxee1->row.pt;
		      	      else
		      	         auxee2 = auxee2->col.pt;
		      } /* while */

		      if (innerprod != 0) {
		      	auxee1 = get_EE();
		      	auxee1->row.ind = j;
		      	auxee1->d_value = innerprod;
		      	auxee1->col.pt = DerQcol[i];
		      	auxee1->tr = MCT_PTH;
		      	DerQcol[i] = auxee1;
		      }
	           } /* for j Adding Qtv * DerX to DerQtt */

		} /* if method sensitivity */

		if (X != NULL)	/* release the entries of X */
			rel_EE(X,VAL_TOFROM);
	} /* for i */

	/* release the entries of Qtv and Pvv*/
	for (j = -Nvan; j < Ntottan; j++) {
		if (Qrow[j] != NULL) {
			rel_EE(Qrow[j],VAL_FROMTO);
			Qrow[j] = NULL;
		}
		if ( Method_sensitivity == VAL_YES )
		   if (DerQrow[j] != NULL) {
			   rel_EE(DerQrow[j],VAL_FROMTO);
			   DerQrow[j] = NULL;
		   }
	}

#ifdef	VERBOSE
	fprintf(stderr,"++++++\t\t\tOperations in procedure 'bsb': %d (+ %d)\n",Narit,Noper);
#endif
} /* backsubst() */

void checkQtt()

/*************************************************************************
At the beginning the matrix is columnwise, at the end it is rowwise.
Double entries (entries in the same column with the same row index)
are collapsed to a single entry.
Self entries in Qtt (diagonal entries) are eliminated.
A count of both is given, together with the # of entries in Q.
*/

{
	EE	*auxee;
	int	ndouble;
	int	nself;
	int	i,row;

	nself = ndouble = Nentries = 0;
	for (i = Ntottan - 1; i >= 0; i--) {
		/* move column i of Qcol */
		while ((auxee = Qcol[i]) != NULL) {
			Qcol[i] = Qcol[i]->col.pt;
			row = auxee->row.ind;
			if (row == i){			/* self entry */
				nself++;
				auxee->col.pt = NULL;
				rel_EE(auxee,VAL_TOFROM);
			} else if (Qrow[row] != NULL && Qrow[row]->col.ind == i) { /* double entry */
				ndouble++;
				Qrow[row]->value += auxee->value;
				auxee->col.pt = NULL;
				rel_EE(auxee,VAL_TOFROM);
			} else {			/* normal entry */
				Nentries++;
				auxee->row.pt = Qrow[row];
				Qrow[row] = auxee;
				Qrow[row]->col.ind = i;
			}
		}
	}
	fprintf(stderr,"After the elimination of vanishing markings:\n");
	if (nself > 0)
		fprintf(stderr,"\t# of self-arcs eliminated:\t%6d\n",nself);
	if (ndouble > 0)
		fprintf(stderr,"\t# of double-arcs compacted:\t%6d\n",ndouble);
	fprintf(stderr,"\t# of remaining arcs:\t\t%6d\n",Nentries);
	Direction = VAL_FROMTO;
} /* checkQtt */

void checkDerQtt()

/*************************************************************************
At the beginning the matrix is columnwise, at the end it is rowwise.
Double entries (entries in the same column with the same row index)
are collapsed to a single entry.
Self entries in DerQtt (diagonal entries) are eliminated.
A count of both is given, together with the # of entries in Q.
*/

{
	EE	*auxee;
	int	ndouble;
	int	nself;
	int	i,row;

	nself = ndouble = Nderentries = 0;
	for (i = Ntottan - 1; i >= 0; i--) {
		/* move column i of DerQcol */
		while ((auxee = DerQcol[i]) != NULL) {
			DerQcol[i] = DerQcol[i]->col.pt;
			row = auxee->row.ind;
			if (row == i){			/* self entry */
				nself++;
				auxee->col.pt = NULL;
				rel_EE(auxee,VAL_TOFROM);
			} else if (DerQrow[row] != NULL && DerQrow[row]->col.ind == i) { /* double entry */
				ndouble++;
				DerQrow[row]->d_value += auxee->d_value;
				auxee->col.pt = NULL;
				rel_EE(auxee,VAL_TOFROM);
			} else {			/* normal entry */
				Nderentries++;
				auxee->row.pt = DerQrow[row];
				DerQrow[row] = auxee;
				DerQrow[row]->col.ind = i;
			}
		}
	}
	fprintf(stderr,"After the elimination of vanishing markings:\n");
	if (nself > 0)
		fprintf(stderr,"\t# of self-arcs eliminated:\t%6d\n",nself);
	if (ndouble > 0)
		fprintf(stderr,"\t# of double-arcs compacted:\t%6d\n",ndouble);
	fprintf(stderr,"\t# of remaining arcs:\t\t%6d\n",Nderentries);
	DerDirection = VAL_FROMTO;
} /* checkDerQtt */

void	row2col(first,last)
int	first,last;

/*************************************************************************
Moves rows from last downto first included into columns, using Qcol and Qrow.
*/

{
	int	row,col;
	EE	*auxee;

	for (row = last; row >= first ; row--) {
		/* move row "row" of Q */
		for (auxee = Qrow[row]; auxee != NULL;) {
			col = auxee->col.ind;
			auxee->col.pt = Qcol[col];
			Qcol[col] = auxee;
			auxee = auxee->row.pt;
			Qcol[col]->row.ind = row;
		}
		Qrow[row] = NULL;
	}
}

void	derrow2dercol(first,last)
int	first,last;

/*************************************************************************
Moves rows from last downto first included into columns, using Qcol and Qrow.
*/

{
	int	row,col;
	EE	*auxee;

	for (row = last; row >= first ; row--) {
		/* move row "row" of Q */
		for (auxee = DerQrow[row]; auxee != NULL;) {
			col = auxee->col.ind;
			auxee->col.pt = DerQcol[col];
			DerQcol[col] = auxee;
			auxee = auxee->row.pt;
			DerQcol[col]->row.ind = row;
		}
		DerQrow[row] = NULL;
	}
	DerDirection = VAL_TOFROM;
}


void	col2row(first,last)
int	first,last;

/*************************************************************************
Moves columns from last downto first included into columns, rows Qcol and Qrow.
*/

{
	int	row,col;
	EE	*auxee;

	for (col = last; col >= first; col--) {
		/* move col "col" of Q */
		for (auxee = Qcol[col]; auxee != NULL;) {
			row = auxee->row.ind;
			auxee->row.pt = Qrow[row];
			Qrow[row] = auxee;
			auxee = auxee->col.pt;
			Qrow[row]->col.ind = col;
		}
		Qcol[col] = NULL;
	}
}

void	dercol2derrow(first,last)
int	first,last;

/*************************************************************************
Moves columns from last downto first included into columns, rows DerQcol
and DerQrow.
*/

{
	int	row,col;
	EE	*auxee;

	for (col = last; col >= first; col--) {
		/* move col "col" of Q */
		for (auxee = DerQcol[col]; auxee != NULL;) {
			row = auxee->row.ind;
			auxee->row.pt = DerQrow[row];
			DerQrow[row] = auxee;
			auxee = auxee->col.pt;
			DerQrow[row]->col.ind = col;
		}
		DerQcol[col] = NULL;
	}
	DerDirection = VAL_FROMTO;
}

void	adjust_prob(dprob,cprob,holding)
double	dprob[],cprob[],holding[];

/*************************************************************************
Input:	The steady state probabilities for the DTMC (in dprob).
	The average holding time (in holding)
Output:	The steady state probabilities for the CTMC (in cprob).
*/

{
	double	sum = 0.0;
	int	i;

	for (i = 0; i < Ntan; i++) {
		cprob[i] = dprob[i] * holding[i];
		sum += cprob[i];
	}
	for (i = Find; i < 0; i++)
		cprob[i] = 0.0;
	for (i = 0; i < Ntan; i++)
		cprob[i] /= sum;
}

void dump(){
	int i; EE *auxee;

	printf("BY ROWS:\n");
	for (i = -Nvan; i < Ntottan; i++) {
		printf("%d	",i);
		for (auxee = Qrow[i]; auxee!=NULL; auxee = auxee->row.pt)
			printf("%d:%lf	",auxee->col.ind,auxee->value);
		printf("\n");
	}
	printf("BY COLUMNS:\n");
	for (i = -Nvan; i < Ntottan; i++) {
		printf("%d	",i);
		for (auxee = Qcol[i]; auxee!=NULL; auxee = auxee->col.pt)
			printf("%d:%lf	",auxee->row.ind,auxee->value);
		printf("\n");
	}
}

void put_in_vector()

/*************************************************************************
Puts the Markov chain in row-pointer-column-index format.
It rotates the Markov chain, if not in the right order (pointed by Qcol).
It normalizes the rows to sum to one (in the vector format only)
if a DTMC is required.
It allocates DProb (possibly) and Cprob, and initiales the right one.
*/

{	int	i,free;
	EE	*auxee;
	double	aux;
extern	int	Direction;
	
	if (Direction == VAL_FROMTO) {
		row2col(-Nvan,Ntottan-1);
		Direction = VAL_TOFROM;
	}
	IP = (int *) new_mem((unsigned int)(sizeof(int) * (Nstates + 1)));
	JA = (int *) new_mem((unsigned int)(sizeof(int) * Nentries));
	A = (double *) new_mem((unsigned int)(sizeof(double) * Nentries));

	aux = 1.0 / Nstates;
	if (IOpt[IOP_MC].val == VAL_DTMC) {
		DProb = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
		CProb = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
		DProb = &DProb[-Find];
		CProb = &CProb[-Find];
		for (i = Find; i < Ntottan; i++)
			DProb[i] = aux;
	} else {
		CProb = (double *) new_mem((unsigned int)
				    (sizeof(double) * Ntottan));

		for (i = Find; i < Ntottan; i++)
			CProb[i] = aux;

		if ( Method_transient == VAL_YES )
		   Cumprob = (double *) new_mem((unsigned int)
					   (sizeof(double) * Ntottan));
	}
	IP = &IP[-Find];
	IP[Find] = 0;
	free = 0;
	for (i = Find; i < Ntottan; i++) {
		for (auxee = Qcol[i]; auxee != NULL; auxee = auxee->col.pt) {
			JA[free] = auxee->row.ind;
			if (IOpt[IOP_MC].val == VAL_DTMC && auxee->row.ind >= 0)
				A[free] = auxee->value*Holding[auxee->row.ind];
			else
				A[free] = auxee->value;
			free++;
		}
		IP[i + 1] = free;
	}

	if ( Method_sensitivity == VAL_YES )
	{
	   if (DerDirection == VAL_FROMTO) {
		   derrow2dercol(-Nvan,Ntottan-1);
		   DerDirection = VAL_TOFROM;
	   }

	   DerCProb = (double *) new_mem((unsigned int)
				    (sizeof(double) * Ntottan));
	   if ( Method_transient == VAL_YES )
	   	DerCumprob = (double *) new_mem((unsigned int)
					   (sizeof(double) * Ntottan));

	   DIP = (int *) new_mem((unsigned int)(sizeof(int) * (Nstates + 1)));
	   DJA = (int *) new_mem((unsigned int)(sizeof(int) * Nderentries));
	   DA = (double *)new_mem((unsigned int)(sizeof(double) * Nderentries));

	   DIP = &DIP[-Find];
	   DIP[Find] = 0;
	   free = 0;
	   for (i = Find; i < Ntottan; i++) {
		   for (auxee = DerQcol[i]; auxee != NULL; auxee = auxee->col.pt) {
			   DJA[free] = auxee->row.ind;
			   DA[free] = auxee->d_value;
			   free++;
		   }
		   DIP[i + 1] = free;
	   }
	}
}

/*
 * initializes the probability values for use in transient analysis.
 * Both Cprob and Cumprob are initialized
 */
void initialize_probs()
{
	int i;

	if ( Method_transient == VAL_YES )
	{
	   for ( i = 0; i < Ntottan; i++ )
	   {
	      CProb[i] = 0.0; Cumprob[i] = 0.0;
	   }
	   CProb[0] = 1.0;

	   if ( Pcan->index < 0 )
	   {
	      for ( i = 0; i < Ntottan; i++ )
		 CProb[i] = Prob_init[i];
	   }
	}

}

void print_dmat()
{
   int index, index2;

   for ( index = 0; index < Ntottan; index++ )
   {
      fprintf(stderr,"index %d: ", index );
      for( index2 = DIP[index]; index2 < DIP[index+1]; index2++ )
      {
	 fprintf(stderr,"%d:%.12g  ", DJA[index2], DA[index2] );
      }
      fprintf(stderr,"\n");
   }

}


int	rg2ctmc()

/*************************************************************************
 It reduces the reachability graph to a tangible to tangible matrix.
 Initially the matrix is columnwise, and so it is at the end.
 RETURNS:
 RES_ERROR or RES_NOERROR.
 */

{
    extern	void	col2row();
    extern	void	dercol2derrow();
	int	i,col;
	EE	*auxee,*auxee1;
	int	first,last;
    
    /* dump(); */
	/* move all the entries to rows */
	col2row(-Nvan,Ntottan-1);
    
	if ( Method_sensitivity == VAL_YES )
	{
        dercol2derrow(-Nvan, Ntottan-1);
	}
    
    /* dump(); */
	Direction = VAL_FROMTO;
	DerDirection = VAL_FROMTO;
	/* perform Gaussian Elimination */
	while (Vloop != NULL) {
		/* triangularize loop pointed by Vloop */
		last = Vloop->first->index;
		first = last - Vloop->length + 1;
		gaussian(Qrow,first,last);
		Vloop = Vloop->pnext;
	}
	/* move the entries V->T and T->T from rows to columns */
	for (i = -Nvan; i < Ntottan; i++) {
		if (Qrow[i] == NULL)		/* empty row */
			continue;
		if (Qrow[i]->col.ind >= 0) {	/* all T,V->T row */
			auxee = Qrow[i];
			Qrow[i] = NULL;
		} else {			/* some T,V->V elements */
			auxee1 = Qrow[i];
			auxee = auxee1->row.pt;
			while (auxee!= NULL && auxee->col.ind < 0) {
				auxee1 = auxee; /* skip T,V->V entries */
				auxee = auxee1->row.pt;
			}
			auxee1->row.pt = NULL;
		}
		while (auxee!= NULL) {
            col = auxee->col.ind;
            auxee->col.pt = Qcol[col];
            Qcol[col] = auxee;
            auxee = auxee->row.pt;
            Qcol[col]->row.ind = i;
		}
	}
	if ( Method_sensitivity == VAL_YES )
	{
        /* move the entries V->T and T->T from rows to columns */
        for (i = -Nvan; i < Ntottan; i++) {
            if (DerQrow[i] == NULL)		/* empty row */
                continue;
            if (DerQrow[i]->col.ind >= 0) {	/* all T,V->T row */
                auxee = DerQrow[i];
                DerQrow[i] = NULL;
            } else {			/* some T,V->V elements */
                auxee1 = DerQrow[i];
                auxee = auxee1->row.pt;
                while (auxee!= NULL && auxee->col.ind < 0) {
                    auxee1 = auxee; /* skip T,V->V entries */
                    auxee = auxee1->row.pt;
                }
                auxee1->row.pt = NULL;
            }
            while (auxee!= NULL) {
                col = auxee->col.ind;
                auxee->col.pt = DerQcol[col];
                DerQcol[col] = auxee;
                auxee = auxee->row.pt;
                DerQcol[col]->row.ind = i;
            }
        }
	}
    
    /* dump(); */
	Direction = VAL_TOFROM;
	DerDirection = VAL_TOFROM;
	backsubst();
    /* dump(); */
	checkQtt();
	if ( Method_sensitivity == VAL_YES )
        checkDerQtt();
    
    
    /* dump(); */
	return(RES_NOERR);
} /* rg2ctmc() */


