/*************************************************************************

FILE:	rglang.c

This file contains the code to read a ".rg" file with the description of
the reachability graph.

FUNCTIONS CONTAINED:
rg_parse()

*/

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"
# include	"var.h"
# include	"lex.h"
# define	RGKEYS	15

static	char	*RGTable[RGKEYS] = {	/* the possible keywords */
		"_t",
		"_a",
		"_v",
		"_l",
		"_nplace",
		"_ntrans",
		"_places",
		"_transitions",
		"_ntanmark",
		"_nabsmark",
		"_nvanmark",
		"_nvanloop",
		"_nentries",
		"_reachset",
		"_reachgraph" };
/* It is used to look up the keywords read by the lexical analyzer. */

int rg_parse(rgfile,full_mark)
FILE	*rgfile;
int	full_mark;	/* full format mark: VAL_YES or VAL_NO */

/*************************************************************************
It parses the reachability graph file.
*/

{
	int	Token,col,row,tr;
	int	pl,tk;
	double	val;
	EE	*auxee;

	/* initialize "Chrd", "Line", "Table", "Nkeyword" */
	Chrd = getc(rgfile);
	Line = 1;
	Table = RGTable;
	Nkeyword = RGKEYS;

for (;;) {
	Token = lex(rgfile);
	/* check if it is the end of the file */
SKLEX:	if (Token == L_EOF)	/* input is not complete: error */
		return(RES_ERROR);
	/* check if token after first keyword of statement is '=' */
	if (lex(rgfile) != L_EQUAL) {
		fprintf(stderr,"ERROR line %d, missing '='\n",Line);
		exit(ER_CODE);
	}
	/* case on the type of statement */
	switch (Token) {
	case R_nplace:
	case R_ntrans:
	case R_ntanmark:
	case R_nabsmark:
	case R_nvanmark:
	case R_nvanloop:
	case R_nentries:
		if (lex(rgfile) != L_NUMBER)
			E_specexp(Table[Token]);
		break;
	case R_places:
	case R_transitions:
/* for the moment the info is not used */
		for (; (Token = lex(rgfile)) == L_NUMBER;) {
			if (lex(rgfile) != L_COLON)
				E_charexp(':');
			if (lex(rgfile) != L_STRING)
				E_specexp("name");
			if (lex(rgfile) != L_SEMICOLON)
				E_charexp(';');
		}
		goto SKLEX;
	case R_reachset:
		for (; (Token = lex(rgfile)) == L_NUMBER;) {
			switch (lex(rgfile)) {
			case R_T:
			case R_A:
			case R_V:
			case R_L:
				/* ignore markings */
				if (full_mark == VAL_YES) {
					for (pl = 0; pl < Nplace; pl++)
						switch (Token = lex(rgfile)) {
						case L_NUMBER:
							if ((tk = Vdouble) < 0 || tk > MAX_TOKEN)
					 			E_valerr("token");
							break;
 						case L_COLON:
							break;
						default:
							E_specexp("token");
						}
				} else {
					while (lex(rgfile) != L_SEMICOLON);
				}
				continue;
			default:
				E_specexp("markingtype");
			}
		}
		goto SKLEX;
	case R_reachgraph:
		while ((Token = lex(rgfile)) == L_NUMBER) {
			row = Vdouble;
			while ((Token = lex(rgfile)) != L_SEMICOLON) {
				if (Token != L_NUMBER)
					 E_specexp("marking");
				col = Vdouble;
				if (lex(rgfile) != L_COLON)
					 E_charexp(':');
				if (lex(rgfile) != L_NUMBER)
					 E_specexp("transition");
				tr = Vdouble;
				if (lex(rgfile) != L_COLON)
					 E_charexp(':');
				if ((Token = lex(rgfile)) != L_NUMBER && Token != L_FLOAT)
					 E_specexp("rate/prob");
				val = Vdouble;
				auxee = get_EE();
				auxee->row.pt = Qrow[row];
				auxee->value = val;
				auxee->d_value = 0.0;
				auxee->tr = tr;
				auxee->col.ind = col;
				Qrow[row] = auxee;
			}
		}
		return(RES_NOERR);
	default:
		E_unknown(Buffer);
	} /* end of switch */

	/* check if the statement is closed by a ';' */
	Token = lex(rgfile);
	if (Token != L_SEMICOLON)
		E_charexp(';');
} /* end of for(;;) */
} /* end of rg_parse() */

