
/*************************************************************************

FILE: spsor.c

*/

# include	<stdio.h>
# include	"port.h"
# include	"const.h"
# include	"lex.h"

extern	char*	new_mem();
extern double fabs();

static	double	*sol_old = NULL;
static	double	*d = NULL;


int	quit_prec(prec)
double	*prec;

/*************************************************************************
This function returns VAL_YES if the user wants to quit, NO otherwise;
the value of the asked precision is changed as well.
*/

{
	double	aux;

	fprintf(stderr,"Current precision: %lg. Please type the new value",*prec);
	fprintf(stderr,"(negative value to stop)\n");
	scanf("%lf",&aux);
	if (aux < 0.0)
		return(VAL_YES);
	(*prec) = aux;
	return(VAL_NO);
}

int	quit_iter(iter)
int	*iter;

/*************************************************************************
This function returns VAL_YES if the user wants to quit, NO otherwise;
the value of the asked number of iterations is changed as well.
*/

{
	int	aux;

	fprintf(stderr,"Current number of iterations: %d. Please type an additional number",*iter);
	fprintf(stderr,"(negative value to stop)\n");
	scanf("%d",&aux);
	if (aux < 0)
		return(VAL_YES);
	(*iter) += aux;
	return(VAL_NO);
}

void sor(ip,ja,a,find,lind,sol,precision,iterations,method)
int	*method;
int	ip[],ja[],find,lind,*iterations;
double	a[],*precision,sol[];
{

/*************************************************************************
Optimal SOR routine for solving linear systems.
This routine is based on the paper by W. Stewart and A. Goyal
in FTCS-16 and on the work of Jon Sjogren in tree-splitting.
The general idea of the routine is to use the SOR method with
modifications to the parameter omega as the computation progresses.

There is an assumption that the Q array is such that it makes sense
to postmultiply the array by the probability vector.
There is also the assumption that this is a rate matrix and so it is not
necessary to store the diagonal in the compressed format.
The diagonal will be computed by this routine.

Input:
ip		an array of indices into the arrays ja and a indicating where
		a sparse row begin and ends.
ja		an array of column indices for each sparse row.
a		the value of the Q array for each entry in the ja array.

sol		initial guess for the probs,
		it is in thie vector that the result is passed back.
precision	tolerance of the solution. The relative difference between two
		iterations should not be greater than this value.
iterations	the number of iterations to perform / performed
method		VAL_SSSOR or VAL_GASEI

Output:
sol		see above under Input.
precision	see above under Input.
iterations	see above under Input.

Local Storage:
sol_old[]	the approximation from the previous iteration.
d[]		array holding the diagonal elements of the Q matrix used
		to scale the Q matrix and save divisions on each iteration.
		The RHS must also be scaled in this case.
delta		the infinity norm of the difference between successive
		approximations, used to compute the new omega parameter.
		delta_1 is the oldest and delta_3 is the youngest.
omega		the overrelaxation parameter.
omega_iter	number of iterations before checking whether to change omega.
*/

	double	omega;
	double	delta_norm;
	double	delta_1;
	double	delta_2;
	double	delta_3;
	double	delta_acc;
	double	epsilon;
	double	vec_norm;
	double	tmp;
	int	c_iter;	/* current number of iterations */
	int	a_iter;	/* asked number of iterations */
	double	c_prec;	/* current precision */
	double	a_prec;	/* asked precision */
	int	compute_new;
	int	omega_iter;
	int	i,j,jstart,jstop;
	int	nstates;
#ifdef	VERBOSE
	int	Noper = 0;
#endif

	omega = 1.0;
	delta_norm = 1.0;
	delta_2 = 10000.0;
	delta_3 = 1000.0;
	delta_acc = 0.0;
	epsilon = 0.10;
	vec_norm = 1.0;
	c_iter = 0;
	a_iter = 0;
	c_prec = 0;
	a_prec = 0;
	compute_new = 0;
	omega_iter = 10;
	nstates = lind - find + 1;
/* Allocate the two arrays, if not already done in a previous call,
and initialize them. */
	if (sol_old == NULL) {
		sol_old = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		d = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		sol_old = &sol_old[-find];
		d = &d[-find];
	}
	for (i = find; i <= lind; i++) {
		sol_old[i] = sol[i];
		d[i] = 0.0;
	}
/* Fill in the diagonal array from the sparse matrix passed in.
Assumes that no diagonal entries are contained in the sparse matrix. */

	for (i = find; i <= lind; i++) {
		for (j = ip[i]; j < ip[i+1]; j++) {
			d[ja[j]] -= a[j];
		}
	}


	a_iter = ((*iterations) == NONE) ? 3000000 : (*iterations);
	a_prec = ((*precision) == NONE) ? 0.0 : (*precision);
	for (;;) {


		if (c_iter >= a_iter) {
			if ((*iterations) == ASK && quit_iter(&a_iter) == NO)
				continue;	/* do not quit */
			if ((*method) == VAL_SSSOR) {
				(*method) = VAL_GASEI;
				c_iter = 0;
				omega = 1.0;
				fprintf(stderr,"ERROR/WARNING: switching from SOR to Gauss-Seidel\n");
				continue;
			}
			break;
		}
		if((c_prec = delta_norm / vec_norm) <= a_prec) {
			if ((*precision) == ASK && quit_prec(&a_prec) == NO)
				continue;	/* do not quit */
			break;
		}
		c_iter++;
/* Main loop, the basic SOR equation. This assumes sorted column indices. */
		for (i = find; i <= lind; i++) {
			jstart = ip[i];
			jstop = ip[i+1];
			tmp = 0.0;
			for (j = jstart; j < jstop; j++) {
#ifdef	VERBOSE
				Noper++;
#endif
				tmp -= a[j] * sol[ja[j]];
			}
			tmp -= (d[i]*sol[i]);


			if (tmp < 1.0e-30 && tmp > -1.0e-30)
				continue;	/* avoid underflow */
			sol[i] += (omega * tmp/d[i]);
			if (sol[i] < 1.0e-30)
				sol[i] = 0.0;	/* avoid underflow */
		}
/* Checks whether the solution vector needs normalization. */
		tmp = 1.0e-25;
		for (i = find; i <= lind; i++)
			if (tmp < sol[i])
				tmp = sol[i];
		if((tmp > 1.0e+10) || (tmp < 1.0e-5)) {
			tmp = 0.0;
			for (i = find; i <= lind; i++)
				tmp += sol[i];
			for (i = find; i <= lind; i++)
				sol[i] /= tmp;
		}
/* Check whether it is time to compute a new omega. */
		if ((*method) == VAL_SSSOR && c_iter == omega_iter) {
			delta_1 = delta_2;
			delta_2 = delta_3;
			delta_3 = delta_acc;
			delta_acc = 0.0;
			compute_new++;
			omega_iter += 10;
/* Every third iteration, check how to adjust omega, based on the
accumulated differences between successive iterations. */
			if (compute_new == 3) {
				compute_new = 0;
				if ((delta_1 <= delta_2) && (delta_2 <= delta_3)) {
/* The iterations are getting worse: move omega back. */
					omega -= 4.0 * epsilon;
					epsilon *= 0.97;
				} else if ((delta_2 >= delta_3) && (delta_2 >= delta_1)) {
					if (delta_1 >= delta_3)
						omega -= epsilon;
					else
						omega -= 3.0 * epsilon;
					epsilon *= 0.89;
				} else if ((delta_2 <= delta_3) && (delta_2 <= delta_1)) {
/*
If the middle delta is the best, then it may be a cusp so
back omega up to just before it and try approaching it from below.
*/
					epsilon *= .5;
					omega -= 3.0 * epsilon;
				} else if ((delta_1 >= delta_2) && (delta_2 >= delta_3)) {
					omega -= epsilon;
					epsilon *= 0.97 ;
				}
			} else {
				omega += epsilon;
/*
If epsilon gets too small, the changes do not have much effect on omega.
Quit and iterate with what we have. It cannot get more optimal.
*/
				if (epsilon < 1.0e-7)
					omega_iter = 3000000;
			}
		}


/*
This code computes the relative difference between successive iterations,
by taking the norm of the difference between successive iterations and
dividing by the norm of the current iterate.
*/
		delta_norm = 0.0;
		vec_norm = 1.0e-25;
		for (i = find; i <= lind; i++) {
			if ((tmp = sol[i] - sol_old[i]) < 0)
				tmp = -tmp;
			if (delta_norm < tmp)
				delta_norm = tmp;
			if (vec_norm < sol[i])
				vec_norm = sol[i];
			sol_old[i] = sol[i];
		}
		delta_acc += (delta_norm/vec_norm);


	} /* end of for(;;) loop */
	/* convergence has been achieved, or too many iterations */
	(*precision) = c_prec;
	(*iterations) = c_iter;
	tmp = 0.0;
	for (i = find; i <= lind; i++)
		tmp += sol[i];
	for (i = find; i <= lind; i++)
		sol[i] /= tmp;
#ifdef	VERBOSE
	fprintf(stderr,"++++++\t\t\tOperations in procedure 'sor': %d\n",Noper);
#endif
} /* end of sor */

void modsor(ip,ja,a,abs,find,lind,sol,neg_rhs,precision,iterations,method)
int	*method;
int	ip[],ja[],find,lind,*iterations;
double	a[],*precision,sol[],neg_rhs[];
double	abs[];	/* if 0.0, the state is to be considered absorbing */
{

/*************************************************************************
neg_rhs		The negative of the right-hand-side
*/

	double	omega;
	double	delta_norm;
	double	delta_1;
	double	delta_2;
	double	delta_3;
	double	delta_acc;
	double	epsilon;
	double	vec_norm;
	double	tmp;
	int	c_iter;	/* current number of iterations */
	int	a_iter;	/* asked number of iterations */
	double	c_prec;	/* current precision */
	double	a_prec;	/* asked precision */
	int	compute_new;
	int	omega_iter;
	int	i,j,jstart,jstop;
	int	nstates;
#ifdef	VERBOSE
	int	Noper = 0;
#endif

/* for (i = 0; i < 10; i++) debug[i] = 1; */
	omega = 1.0;
	delta_norm = 1.0;
	delta_2 = 10000.0;
	delta_3 = 1000.0;
	delta_acc = 0.0;
	epsilon = 0.10;
	vec_norm = 1.0;
	c_iter = 0;
	a_iter = 0;
	c_prec = 0;
	a_prec = 0;
	compute_new = 0;
	omega_iter = 10;
	nstates = lind - find + 1;
/* Allocate the two arrays, if not already done in a previous call,
and initialize them. */
	if (sol_old == NULL) {
		sol_old = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		sol_old = &sol_old[-find];
		d = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		d = &d[-find];
	}
	for (i = find; i <= lind; i++) {
		if (abs[i] == 0.0)
			continue;
		sol_old[i] = sol[i];
		d[i] = 0.0;
	}
/* Fill in the diagonal array from the sparse matrix passed in.
Assumes that no diagonal entries are contained in the sparse matrix. */


 	for (i = find; i <= lind; i++) {
		for (j = ip[i]; j < ip[i+1]; j++) {
			if (abs[ja[j]] == 0.0)
				continue;
			d[ja[j]] -= a[j];
		}
	}

	a_iter = ((*iterations) == NONE) ? 3000000 : (*iterations);
	a_prec = ((*precision) == NONE) ? 0.0 : (*precision);
	for (;;) {


		if (c_iter >= a_iter) {
			if ((*iterations) == ASK && quit_iter(&a_iter) == NO)
				continue;	/* do not quit */
			if ((*method) == VAL_SSSOR) {
				(*method) = VAL_GASEI;
				c_iter = 0;
				omega = 1.0;
				fprintf(stderr,"ERROR/WARNING: switching from SOR to Gauss-Seidel\n");
				continue;
			}
			break;
		}
		if((c_prec = delta_norm / vec_norm) <= a_prec) {
			if ((*precision) == ASK && quit_prec(&a_prec) == NO)
				continue;	/* do not quit */
			break;
		}
		c_iter++;
/* Main loop, the basic SOR equation. This assumes sorted column indices. */
		for (i = find; i <= lind; i++) {
			if (abs[i] == 0.0)
				continue;
			jstart = ip[i];
			jstop = ip[i+1];
			tmp = -neg_rhs[i];
			for (j = jstart; j < jstop; j++) {
				if (abs[ja[j]] == 0.0)
					continue;
#ifdef	VERBOSE
				Noper++;
#endif
				tmp -= a[j] * sol[ja[j]];
			}
			tmp -= (d[i]*sol[i]);


			if (tmp < 1.0e-30 && tmp > -1.0e-30)
				continue;	/* avoid underflow */
			sol[i] += (omega * tmp/d[i]);
			if (sol[i] < 1.0e-30)
				sol[i] = 0.0;	/* avoid underflow */
		}
/* Checks whether the solution vector needs normalization.
		tmp = 1.0e-25;
		for (i = find; i <= lind; i++) {
			if (abs[i] == 0.0)
				continue;
			if (tmp < sol[i])
				tmp = sol[i];
		}
		if((tmp > 1.0e+25) || (tmp < 1.0e-10)) {
			tmp = 0.0;
			for (i = find; i <= lind; i++) {
				if (abs[i] == 0.0)
					continue;
				tmp += sol[i];
			}
			for (i = find; i <= lind; i++) {
				if (abs[i] == 0.0)
					continue;
				sol[i] /= tmp;
			}
		} */
/* Check whether it is time to compute a new omega. */
		if ((*method) == VAL_SSSOR && c_iter == omega_iter) {
			delta_1 = delta_2;
			delta_2 = delta_3;
			delta_3 = delta_acc;
			delta_acc = 0.0;
			compute_new++;
			omega_iter += 10;
/* Every third iteration, check how to adjust omega, based on the
accumulated differences between successive iterations. */
			if (compute_new == 3) {
				compute_new = 0;
				if ((delta_1 <= delta_2) && (delta_2 <= delta_3)) {
/* The iterations are getting worse: move omega back. */
					omega -= 4.0 * epsilon;
					epsilon *= 0.97;
				} else if ((delta_2 >= delta_3) && (delta_2 >= delta_1)) {
					if (delta_1 >= delta_3)
						omega -= epsilon;
					else
						omega -= 3.0 * epsilon;
					epsilon *= 0.89;
				} else if ((delta_2 <= delta_3) && (delta_2 <= delta_1)) {
/*
If the middle delta is the best, then it may be a cusp so
back omega up to just before it and try approaching it from below.
*/
					epsilon *= .5;
					omega -= 3.0 * epsilon;
				} else if ((delta_1 >= delta_2) && (delta_2 >= delta_3)) {
					omega -= epsilon;
					epsilon *= 0.97 ;
				}
			} else {
				omega += epsilon;
/*
If epsilon gets too small, the changes do not have much effect on omega.
Quit and iterate with what we have. It cannot get more optimal.
*/
				if (epsilon < 1.0e-7)
					omega_iter = 3000000;
			}
		}


/*
This code computes the relative difference between successive iterations,
by taking the norm of the difference between successive iterations and
dividing by the norm of the current iterate.
*/
		delta_norm = 0.0;
		vec_norm = 1.0e-25;
		for (i = find; i <= lind; i++) {
			if (abs[i] == 0.0)
				continue;
			if ((tmp = sol[i] - sol_old[i]) < 0)
				tmp = -tmp;
			if (delta_norm < tmp)
				delta_norm = tmp;
			if (vec_norm < sol[i])
				vec_norm = sol[i];
			sol_old[i] = sol[i];
		}
		delta_acc += (delta_norm/vec_norm);


	} /* end of for(;;) loop */
	/* convergence has been achieved, or too many iterations */
	(*precision) = c_prec;
	(*iterations) = c_iter;
/*	tmp = 0.0;
	for (i = find; i <= lind; i++) {
		if (abs[i] == 0.0)
			continue;
		tmp += sol[i];
	}
	for (i = find; i <= lind; i++) {
		if (abs[i] == 0.0)
			continue;
		sol[i] /= tmp;
	} */
#ifdef	VERBOSE
	fprintf(stderr,"++++++\t\t\tOperations in procedure 'modsor': %d\n",Noper);
#endif
} /* end of sor */

void newsor(ip,ja,a,find,lind,sol,rhs,prob,precision,iterations,method)
int	*method;
int	ip[],ja[],find,lind,*iterations;
double	a[],*precision,sol[],rhs[],prob[];
{

/*************************************************************************
rhs		The negative of the right-hand-side
*/

	double	omega;
	double	delta_norm;
	double	delta_1;
	double	delta_2;
	double	delta_3;
	double	delta_acc;
	double	epsilon;
	double	vec_norm;
	double	tmp;
	int	c_iter;	/* current number of iterations */
	int	a_iter;	/* asked number of iterations */
	double	c_prec;	/* current precision */
	double	a_prec;	/* asked precision */
	int	compute_new;
	int	omega_iter;
	int	i,j,jstart,jstop;
	int	nstates;
//	int	tempint;
#ifdef	VERBOSE
	int	Noper = 0;
#endif

/* for (i = 0; i < 10; i++) debug[i] = 1; */
	omega = 1.0;
	delta_norm = 1.0;
	delta_2 = 10000.0;
	delta_3 = 1000.0;
	delta_acc = 0.0;
	epsilon = 0.10;
	vec_norm = 1.0;
	c_iter = 0;
	a_iter = 0;
	c_prec = 0;
	a_prec = 0;
	compute_new = 0;
	omega_iter = 10;
	nstates = lind - find + 1;
/* Allocate the two arrays, if not already done in a previous call,
and initialize them. */
	if (sol_old == NULL) {
		sol_old = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		sol_old = &sol_old[-find];
		d = (double *) new_mem((unsigned int)(sizeof(double) * nstates));
		d = &d[-find];
	}
	for (i = find; i <= lind; i++) {
		sol_old[i] = sol[i];
		d[i] = 0.0;
	}

/* Fill in the diagonal array from the sparse matrix passed in.
Assumes that no diagonal entries are contained in the sparse matrix. */


 	for (i = find; i <= lind; i++) {
		for (j = ip[i]; j < ip[i+1]; j++) {
			d[ja[j]] -= a[j];
		}
	}

	a_iter = ((*iterations) == NONE) ? 3000000 : (*iterations);
	a_prec = ((*precision) == NONE) ? 0.0 : (*precision);
	for (;;) {


		if (c_iter >= a_iter) {
			if ((*iterations) == ASK && quit_iter(&a_iter) == NO)
				continue;	/* do not quit */
			if ((*method) == VAL_SSSOR) {
				(*method) = VAL_GASEI;
				c_iter = 0;
				omega = 1.0;
				fprintf(stderr,"ERROR/WARNING: switching from SOR to Gauss-Seidel\n");
				continue;
			}
			break;
		}
		if((c_prec = delta_norm / vec_norm) <= a_prec) {
			if ((*precision) == ASK && quit_prec(&a_prec) == NO)
				continue;	/* do not quit */
			break;
		}
		c_iter++;
/* Main loop, the basic SOR equation. This assumes sorted column indices. */
		for (i = find; i <= lind; i++) {
			jstart = ip[i];
			jstop = ip[i+1];
			tmp = rhs[i];
			for (j = jstart; j < jstop; j++) {
#ifdef	VERBOSE
				Noper++;
#endif
				tmp -= a[j] * sol[ja[j]];
			}
			tmp -= (d[i]*sol[i]);
			sol[i] += (omega * tmp / d[i] );

		}
/* Checks whether the solution vector needs normalization. */
		if ( c_iter % 10 == 0 )
		{
		tmp = 0.0;
		for ( i = find; i <= lind; i++ )
			tmp += sol[i];

		for( i = find; i <= lind; i++ )
			sol[i] -= tmp * prob[i];
		}

/* Check whether it is time to compute a new omega. */
		if ((*method) == VAL_SSSOR && c_iter == omega_iter) {
			delta_1 = delta_2;
			delta_2 = delta_3;
			delta_3 = delta_acc;
			delta_acc = 0.0;
			compute_new++;
			omega_iter += 10;
/* Every third iteration, check how to adjust omega, based on the
accumulated differences between successive iterations. */
			if (compute_new == 3) {
				compute_new = 0;
				if ((delta_1 <= delta_2) && (delta_2 <= delta_3)) {
/* The iterations are getting worse: move omega back. */
					omega -= 4.0 * epsilon;
					epsilon *= 0.97;
				} else if ((delta_2 >= delta_3) && (delta_2 >= delta_1)) {
					if (delta_1 >= delta_3)
						omega -= epsilon;
					else
						omega -= 3.0 * epsilon;
					epsilon *= 0.89;
				} else if ((delta_2 <= delta_3) && (delta_2 <= delta_1)) {
/*
If the middle delta is the best, then it may be a cusp so
back omega up to just before it and try approaching it from below.
*/
					epsilon *= .5;
					omega -= 3.0 * epsilon;
				} else if ((delta_1 >= delta_2) && (delta_2 >= delta_3)) {
					omega -= epsilon;
					epsilon *= 0.97 ;
				}
			} else {
				omega += epsilon;
/*
If epsilon gets too small, the changes do not have much effect on omega.
Quit and iterate with what we have. It cannot get more optimal.
*/
				if (epsilon < 1.0e-7)
					omega_iter = 3000000;
			}
		}


/*
This code computes the relative difference between successive iterations,
by taking the norm of the difference between successive iterations and
dividing by the norm of the current iterate.
*/
		delta_norm = 0.0;
		vec_norm = 1.0e-25;
		for (i = find; i <= lind; i++) {
			if ((tmp = sol[i] - sol_old[i]) < 0)
				tmp = -tmp;
			if (delta_norm < tmp)
				delta_norm = tmp;
			if (vec_norm < sol[i])
				vec_norm = sol[i];
			sol_old[i] = sol[i];
		}
		delta_acc += (delta_norm/vec_norm);


	} /* end of for(;;) loop */

	/* final normalizatio step */
	tmp = 0.0;
	for ( i = find; i <= lind; i++ )
		tmp += sol[i];

	for( i = find; i <= lind; i++ )
		sol[i] -= tmp * prob[i];

	/* convergence has been achieved, or too many iterations */
	(*precision) = c_prec;
	(*iterations) = c_iter;

#ifdef	VERBOSE
	fprintf(stderr,"++++++\t\t\tOperations in procedure 'modsor': %d\n",Noper);
#endif
} /* end of newsor */
