
/*************************************************************************

FILE: tsufsens.c

*/

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"const.h"
# include	"lex.h"
# define max(x,y) (((x) > (y)) ? (x) : (y))

extern	char*	new_mem();
double fabs();
double exp();
double log();


void unif_sens(IP,JA,A,Nstates,Nentries,DIP,DJA,DA,Nderentries, Prob,Cumprob,
				  Derprob,DerCumprob,Precision,time,Method)
int	*Method;
int	IP[],JA[],Nentries,Nstates;
int	DIP[],DJA[],Nderentries;
double	A[],DA[],*Precision,Prob[],Cumprob[],Derprob[],DerCumprob[];
double  time;
{

/*************************************************************************

Input:
Order		M_TOFROM or M_FROMTO
IP		an array of indices into the arrays JA and A indicating where
		a sparse row begin and ends.
JA		an array of column indices for each sparse row.
A		the value of the Q array for each entry in the JA array.

Prob		vector representing the unknown of the equation and it is in
		thie vector that the result is passed back.
Precision	tolerance of the solution. The relative difference between two
		iterations should not be greater than this value.
time		the time at which the probability needs to be evaluated

Output:
Prob		see above under Input.

Local Storage:
Prob_old[]	the approximation from the previous iteration.
d[]		array holding the diagonal elements of the Q matrix used
		to scale the Q matrix and save divisions on each iteration.
		The RHS must also be scaled in this case.
*/

   static double *Prob_old = NULL;
   //static double *Derprob_old = NULL;
   double *d;
   double *dd;
   double tmp;
   double tmp2;
   int i,j,jstart;
   //int jstop;
   double q;
   double *A1;
   double *d1;
   double *DA1;
   double *dd1;
   double *s, *r;
   double *r_old;
   double *ds, *dr;
   double *dr_old;
   double qt;
   double sumqt;
   double tmpti;
   double save_tmpti;
   double expqt;
   double minqt;
   double log_qt;
   double p_term;
   double rcum;
   double ri;
   double *L;
   double *P;
   double *P0;
   double *DL;
   double *DP;
   double *DP0;
   int k;
   int l;
   double epsilon_2;

#ifdef	VERBOSE
	int	Noper = 0;
#endif

   /*
   if (Order != M_TOFROM)
   {
      fprintf(stderr,"Sorry, only TOFROM order is allowed\n");
      exit(ER_CODE);
   }
   */

/* Allocate the two arrays, if not already done in a previous call,
and initialize them. */
   if (Prob_old == NULL)
   {
      Prob_old = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      d = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      A1 = (double *) new_mem((unsigned int)(sizeof(double) * Nentries));
      d1 = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      dd = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      DA1 = (double *) new_mem((unsigned int)(sizeof(double) * Nderentries));
      dd1 = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      s = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      r = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      ds = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      dr = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      L = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      P = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      P0 = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      DL = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      DP = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      DP0 = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      r_old = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
      dr_old = (double *) new_mem((unsigned int)(sizeof(double) * Nstates));
   }

   for (i = 0; i < Nstates; i++)
   {
      Prob_old[i] = Prob[i];
      P0[i] = Prob[i];
      DP0[i] = 0.0;
      d[i] = 0.0;
      dd[i] = 0.0;
   }

/* Fill in the diagonal array from the sparse matrix passed in.
Assumes that no diagonal entries are contained in the sparse matrix. */

   for (i = 0; i < IP[Nstates]; i++)
      d[JA[i]] -= A[i];

   for (i = 0; i < DIP[Nstates]; i++)
      dd[DJA[i]] -= DA[i];

   /* debugging output 
   for ( i = 0; i < Nstates; i++ )
   {
      fprintf(stdout,"%d   ", i );
      for ( j = IP[i]; j < IP[i+1]; j++ )
	 fprintf(stdout,"%d:%f  ", JA[j], A[j] );
      fprintf(stdout,"%d:%f\n",i, d[i]);
   }
   fprintf(stdout,"\n");
 debugging output */

/* select the value of q as the largest diagonal value in the matrix */
   q = 0;
   for ( i = 0; i < Nstates; i++ )
   {
      tmp = max(q, fabs(d[i]));
      q = tmp;
   }

   /* just testing if q > max|-qii| has any effect
   q *= 1.2;
    commented out for now */

   for ( i = 0; i < IP[Nstates]; i++ )
      A1[i] = A[i] / q;

   for ( i = 0; i < DIP[Nstates]; i++ )
      DA1[i] = DA[i] / q;

   for ( i = 0; i < Nstates; i++ )
   {
      d1[i] = d[i] / q + 1;
      dd1[i] = dd[i] / q;
   }

   /* debugging output 
   fprintf(stdout,"THE DTMC Q* MATRIX\n");
   fprintf(stdout, " the value of q: %f\n", q );
   for ( i = 0; i < Nstates; i++ )
   {
      fprintf(stdout,"%d   ", i );
      for ( j = IP[i]; j < IP[i+1]; j++ )
	 fprintf(stdout,"%d:%f  ", JA[j], A1[j] );
      fprintf(stdout,"%d:%f\n",i, d1[i]);
   }
   fprintf(stdout,"\n");
    end of debugging */

   /* starting of the uniformization algorithm */
   l = 0;
   epsilon_2 = *Precision / 2.0;
   qt = q * time;
   minqt = -qt;
   log_qt = log(qt);
   expqt = exp(minqt);

   /* debugging
   fprintf(stdout,"logqt %.12f expqt %.12f\n", log_qt, expqt);
   fprintf(stdout," time %.12f qt %.12f exp(-qt)  %.12f\n", time, qt, expqt );
    debugging */

   tmpti = minqt;
   p_term = exp(tmpti);
   sumqt = 0.0;

   /* debugging
      fprintf(stdout, "P_term %.12f sumqt %.12f\n", p_term, sumqt );
    debugging */

   /* initialize r */
   for ( i = 0; i < Nstates; i++ )
   {
      r_old[i] = r[i] = Prob[i];
      dr_old[i] = dr[i] = 0.0;

   }

   while ((sumqt + p_term < epsilon_2) || (p_term < UNDFLOW))
   {
      l++;
      sumqt += p_term;
      tmpti += ( log_qt - log((double)l) );
      p_term = exp(tmpti);
      for( j = 0; j < Nstates; j++ )
      {
	 r[j] = r_old[j] * d1[j];
	 dr[j] = dr_old[j] * d1[j];
	 dr[j] += r_old[j] * dd1[j];
	 for( jstart = IP[j]; jstart < IP[j+1]; jstart++ )
	 {
	    r[j] += r_old[JA[jstart]]*A1[jstart];
	    dr[j] += dr_old[JA[jstart]]*A1[jstart];
	 }
	 for( jstart = DIP[j]; jstart < DIP[j+1]; jstart++ )
	 {
	    dr[j] += r_old[DJA[jstart]]*DA1[jstart];
	 }
      } /* for j */
      for ( i = 0; i < Nstates; i++ )
      {
         r_old[i] = r[i];
         dr_old[i] = dr[i];
      }

   } /* while sumqt */

   k = l;
   save_tmpti = tmpti;

   if ( l > 0 )
   {
      while ( (epsilon_2 < ( 1.0 - sumqt)) && (p_term > UNDFLOW) )
      {
         k++;
         tmpti += ( log_qt - log((double)k) );
         p_term = exp(tmpti);
         sumqt += p_term;
   
         /* debugging 
         fprintf(stdout, "P_term %.12f sumqt %.12f\n", p_term, sumqt );
          end of debugging */
   
      } /* end while */
   } /* if l > 0 */
   else
   {
      while ( (*Precision < ( 1.0 - sumqt)) && (p_term > UNDFLOW) )
      {
         k++;
         tmpti += ( log_qt - log((double)k) );
         p_term = exp(tmpti);
         sumqt += p_term;
   
         /* debugging 
         fprintf(stdout, "P_term %.12f sumqt %.12f\n", p_term, sumqt );
          end of debugging */
   
      } /* end while */
   }

   /* debugging */
   fprintf(stdout,"The value of l:%d and k:%d\n",l, k );
   /*  debugging */

   /* initialize s */
   for ( i = 0; i < Nstates; i++ )
   {
      s[i] = 0;
      ds[i] = 0;

      /* debugging
      fprintf(stdout, "Prob[%d] = %.12f\n", i, Prob[i] );
      end of debugging */

   }

   tmpti = save_tmpti;
   p_term = exp(tmpti);
   /* start the computation of the probabilities */
   for ( i = l; i <= (k+1) ; i++ )
   {
      for ( j = 0; j < Nstates; j++ )
      {
	 s[j] += r[j]*p_term;
	 ds[j] += dr[j]*p_term;
	 r_old[j] = r[j];
	 dr_old[j] = dr[j];

	 /* debugging
	 fprintf(stdout,"r[%d] = %.12g\n", j, r[j] );
	  debugging */
      }

      tmpti += ( log_qt - log((double)(i+1)) );
      p_term = exp(tmpti);
      for( j = 0; j < Nstates; j++ )
      {
	 r[j] = r_old[j] * d1[j];
	 dr[j] = dr_old[j] * d1[j];
	 dr[j] += r_old[j] * dd1[j];
	 for( jstart = IP[j]; jstart < IP[j+1]; jstart++ )
	 {
	    r[j] += r_old[JA[jstart]]*A1[jstart];
	    dr[j] += dr_old[JA[jstart]]*A1[jstart];
	 }
	 for( jstart = DIP[j]; jstart < DIP[j+1]; jstart++ )
	 {
	    dr[j] += r_old[DJA[jstart]]*DA1[jstart];
	 }
      } /* for j */

   } /* for i */

   /* debugging 
   tmp = 0;
    end of debugging */
   tmp2 = 0.0;

   for ( i = 0; i < Nstates; i++ )
   {
      Prob[i] = s[i];
      Derprob[i] = ds[i];

   /* debugging
      tmp += Prob[i];
      fprintf(stdout, "Prob[%d] = %.12f\n", i, Prob[i] );
     end of debugging */
      tmp2 += Derprob[i];

   }

   /* debugging 
   fprintf(stdout, "Sum of probabilities = %.12f\n", tmp );
    end of debugging */
   fprintf(stdout, "Sum of derivatives of probabilities = %.12f\n", tmp2 );
   
   /* The following code is for cumulative probability analysis */

   k = 1;
   tmpti = -qt;
   p_term = exp(tmpti);
   sumqt = p_term;
   while ( *Precision < time*( 1.0 - sumqt) )
   {
      tmpti += ( log_qt - log((double)k));
      p_term = exp(tmpti);
      sumqt += p_term;
      k++;
   } /* end while */

   /* debugging  */
   fprintf(stdout,"Cumulative Unif. k = %d\n", k);
   /* end of debugging */

   for ( i = 0; i < Nstates; i++ )
   {
      L[i] = P0[i] * ( 1.0 - expqt);
      DL[i] = 0.0;
      P[i] = P0[i];
      DP[i] = 0.0;
   }
   rcum = 1.0 - expqt;
   tmpti = -qt;
   ri = exp(tmpti);
   for( j = 1; j <= k; j++ )
   {
      for( i = 0; i < Nstates; i++ )
      {
	 P0[i] = P[i];
	 DP0[i] = DP[i];
      }

      tmp = 0;
      for( i = 0; i < Nstates; i++ )
      {
	 P[i] = P0[i] * d1[i];
	 DP[i] = DP0[i] * d1[i];
	 DP[i] += P0[i] * dd1[i];
	 for( jstart = IP[i]; jstart < IP[i+1]; jstart++ )
	 {
	    P[i] += P0[JA[jstart]]*A1[jstart];
	    DP[i] += DP0[JA[jstart]]*A1[jstart];
         }
	 for( jstart = DIP[i]; jstart < DIP[i+1]; jstart++ )
	 {
	    DP[i] += P0[DJA[jstart]]*DA1[jstart];
         }

	 /* debugging 
	 tmp += P[i];
	 fprintf(stdout,"P[%d] = %.12f\n", i, P[i] );
	  end of debugging */

      } /* for i */

	 /* debugging 
      fprintf(stdout,"Total Prob %.12f\n", tmp );
	  end of debugging */

      tmpti += ( log_qt - log((double)j) );
      ri = exp(tmpti);
      rcum -= ri;

      /* debugging 
      fprintf(stdout," ri = %.12g  rcum = %.12g\n", ri, rcum );
       end of debugging */

      for( i = 0; i < Nstates; i++ )
      {
	 L[i] += rcum * P[i];
	 DL[i] += rcum * DP[i];
      }
   }/* for j */

   /* debugging
   tmp = 0.0;
   end of debugging */
   tmp2 = 0.0;

   for ( i = 0; i < Nstates; i++ )
   {
      Cumprob[i] = 0.0;
      DerCumprob[i] = 0.0;
      Cumprob[i] = L[i] / q;
      DerCumprob[i] = DL[i] / q;

   /* debugging
      tmp += Cumprob[i];
   end of debugging */
      tmp2 += DerCumprob[i];

   }

   /* debugging
   fprintf(stdout, "Total cumulative Prob %.12f\n", tmp );
   end of debugging */
   fprintf(stdout, "Total derivative cumulative Prob %.12f\n", tmp2 );

   free(Prob_old);
   free(d);
   free(dd);
   free(A1);
   free(DA1);
   free(d1);
   free(dd1);
   free(s);
   free(ds);
   free(r);
   free(dr);
   free(L);
   free(P);
   free(P0);
   free(DL);
   free(DP);
   free(DP0);
   free(r_old);
   free(dr_old);

   Prob_old = d = A1 = d1 = s = r = L = P = P0 = r_old = NULL;
   DL = DP = DP0 = dd = DA1 = dd1 = ds = dr = dr_old = NULL;

} /* end of uniform */
