
/*************************************************************************
FILE:	type.h
*/

/* VE: variable element
used to store a pointer to a function returning an integer */

typedef int (* VE)();

/* VA: variable array
used to store an array of pointers to functions returning integers */

typedef VE	*VA;

/* ST: symbol type
used to store a symbol (place, transition, parameter) in the symbol table */

typedef struct st {

short	numb;		/* index in the Arrplace, Arrtrans, or Arrparam */
char	type;		/* SYM_PLACE, SYM_TRANS, SYM_PARAM, SYM_EMPTY */
char	*symb;		/* pointer to the symbol string */

}	ST;

/* PP: place position
used to store the number of tokens in a place, used in an array */

typedef	char	PP;

/* PL: place type.
Used to store a place */

typedef struct pl {

char	*name;		/* name of the place */
short	n1;		/* total number on 1st dimension (>= 1) */
short	n2;		/* total number on 2nd dimension (>= 1) */
short	i1;		/* index on 1st dimension (1..n1) */
short	i2;		/* index on 2nd dimension (1..n2) */
PP	initial;	/* initial number of tokens */
short	num;		/* number of the place, user order */
struct	pl	*pnext;	/* pointer to the next place */

}	PL;

/* TR: transition type.
Used to store a transition */

typedef struct tr {

char	*name;		/* name of the transition */
short	n1;		/* total number on 1st dimension (>= 1) */
short	n2;		/* total number on 2nd dimension (>= 1) */
short	i1;		/* index on 1st dimension (1..n1) */
short	i2;		/* index on 2nd dimension (1..n2) */
short	num;		/* number of the transition, user order */
short	prio;		/* priority */
char	type;		/* DIS_x */
char	dep;		/* DEP_x */
int	pl;		/* index of place, used if DEP_PL */
PP	*inp;		/* constant input array */
VA	vin;		/* marking dependent input array */
PP	*out;		/* constant output array */
VA	vou;		/* marking dependent output array */
PP	*inh;		/* constant inhibition array */
VA	vih;		/* marking dependent inhibition array */
int	(*enabl)();	/* pointer to the enabling function */
double	val;		/* constant prob/rate value, if any */
double	dval;		/* constant derivative prob/rate value, if any */
double	(*fun)();	/* pointer to the prob/rate function */
double	(*dfun)();	/* pointer to the derivative prob/rate function */
struct	tr	*pnext;	/* pointer to the next transition */

}	TR;


/* EE: entry element type.
Used to store graph arcs, and subsequentely the transition matrix */

typedef struct	ee {

short	tr;		/* number of the PN-transition (user order) */
double	value;		/* rate or probability */
double	d_value;	/* derivative of rate or probability */
union {
int	ind;		/* index of the row */
struct	ee	*pt;	/* pointer to the next element in the row */
}	row;
union {
int	ind;		/* index of the column */
struct	ee	*pt;	/* pointer to the next element in the column */
struct	me	*mk;	/* "mk->index" is equivalent to "ind" */
}	col;

}	EE;


/* ME: stored marking element type.
Used to store the AVL tree, and the pointers to the graph arcs (EE entries) */

typedef struct	me {

int	index;		/* marking index */
struct	me	*pcan;	/* pointer to next marking in canonical order */
struct	me	*plr[2];/* pointers to two children in AVL tree */
char	bal;		/* LT,GT,EQ: balance of the AVL node */
char	type;		/* TAN_M,ABS_M,REP_M,VAN_M,VLO_M,ALO_M: marking type */
PP	*mk;		/* array representing number of tokens in each place */
EE	*prow;		/* pointer to the row of entries */

}	ME;



/* LL: loop list type.
Used to store loops of vanishing markings during the reach set construction */

typedef struct	ll {

struct	me	*first;	/* pointer to the first marking of the loop */
int	length;		/* number of markings in the loop */
struct	ll	*pnext;	/* pointer to the next loop */

}	LL;


/* ID: integer option description.
used to store the integer options */

typedef struct	id {

char	*str;		/* pointer to the name of the option */
char	leg[5];		/* possible values, the last one is 0 */
int	val;		/* actual value */

}	ID;


/* FD: floating point option description.
used to store the floating point options */

typedef struct	fd {

char	*str;		/* pointer to the name of the option */
double	val;		/* actual value */

}	FD;
