
/*LINTLIBRARY*/

/*************************************************************************

FILE:	utility.c

Utility functions are contained here.

FUNCTIONS CONTAINED:
	getanswer()
	new_mem()

*/

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"

//extern char	*malloc();

int getanswer()

/*************************************************************************
Return VAL_YES if user types an 'y' or a 'Y' , VAL_NO if he types an 'n' or 'N'.
All other characters are simply ignored.
*/

{
	int	ch;

	for (;;) {
		ch = getchar();
		if (ch == 'y' || ch == 'Y')
			return(VAL_YES);
		if (ch == 'n' || ch == 'N')
			return(VAL_NO);
	}
} /* getanswer() */

char	*new_mem(size)
unsigned int	size;

/*************************************************************************
Memory management function, to allocate clean memory.

size		number of bytes to be allocated

RETURNS:	the address of the first of the allocated bytes
*/
{
//extern	char	*malloc();
	char	*pt;
	int	i;

	if ((pt = malloc(size)) == NULL) {
		fprintf(stderr,"ABORT: not enough memory\n");
		exit(ER_CODE);
	}
	for (i = 0; i < size; i++)
		pt[i] = '\0';
	return(pt);
}

# define	ALL_SIZE	1000
/* number of entries allocated at the same time */
# define	ALL1_SIZE	999
/* ALL_SIZE - 1 */

static	EE	*freeEE = NULL;
static	LL	*freeLL = NULL;

EE	*get_EE()

/*************************************************************************
Memory management functions for the dynamically allocated structures.
The "get_XX" functions allocate one element of type "XX".

RETURNS:	the address of the allocated structure
*/

{
	EE	*auxee;
	int	i;

	if (freeEE == NULL) {
		freeEE = (EE *) new_mem((unsigned int)(sizeof(EE) * ALL_SIZE));
		for (i = 0; i < ALL1_SIZE; i++)
			(&freeEE[i])->col.pt = &freeEE[i+1];
		(&freeEE[ALL1_SIZE])->col.pt = NULL;
	}
	auxee = freeEE;
	freeEE = freeEE->col.pt;
	auxee->col.pt = NULL;
	return(auxee);
} /* get_EE() */

ME	*get_ME(length)
unsigned length;
{
	ME	*auxme;
static unsigned nME = 0;	/* count of the number of ME free */
static unsigned nPP = 0;	/* count of the number of PlacePositions free */
static	ME	*freeME;	/* array of free ME */
static	PP	*freePP;	/* array of free PP */

	if (nME == 0) {
		nME = ALL_SIZE;
		freeME = (ME *) new_mem((unsigned int)(sizeof(ME) * nME));
	}
	if (nPP < length) {	/* throw away the small piece */
		nPP = (length < ALL_SIZE) ? ALL_SIZE : length;
/*###113 [lint] new_mem arg. 1 used inconsistently utility.c(51) :: utility.c(113)%%%*/
		freePP = (char *) new_mem((unsigned int)(sizeof(PP) * nPP));
	}
	freeME->mk = freePP;
	auxme = freeME;
	freeME++;
	freePP += length;
	nME--;
	nPP -= length;
	return(auxme);
} /* get_ME() */

LL	*get_LL()
{
	LL	*auxll;

	if (freeLL == NULL)
		auxll = (LL *) new_mem((unsigned int)(sizeof(LL)));
	else {
		auxll = freeLL;
		freeLL = freeLL->pnext;
	}
	return(auxll);
} /* get_LL() */

char	*get_SY(length)
unsigned length;
{
static	char	*freeSY = NULL;
	char	*auxsy;
static unsigned nSY = 0;	/* count of the number of chars free */

	if (nSY == 0 || nSY < length) {
		nSY = (length < ALL_SIZE) ? ALL_SIZE : length;
/*###146 [lint] new_mem arg. 1 used inconsistently utility.c(51) :: utility.c(146)%%%*/
		freeSY = (char *) new_mem((unsigned int)(sizeof(char) * nSY));
	}
	auxsy = freeSY;
	freeSY += length;
	nSY -= length;
	return(auxsy);
} /* get_SY() */

void rel_EE(first,order)
EE	*first;	 /* the first element of the chain */
int	order;		/* VAL_FROMTO if row chain, VAL_TOFROM otherwise */

/*************************************************************************
The "rel_XX" functions deallocate all the list pointed by the input param.

RETURNS:	nothing
*/

{
	EE	*auxee;

	auxee = first;
	if (order == VAL_FROMTO)
		while ((auxee->col.pt = auxee->row.pt) != NULL)
			auxee = auxee->col.pt;
	else
		while (auxee->col.pt != NULL)
			auxee = auxee->col.pt;
	auxee->col.pt = freeEE;
	freeEE = first;
}

void rel_LL(first)
LL	*first;	 /* the first element of the chain */
{
	LL	*auxll;

	auxll = first;
	while (auxll->pnext != NULL)
		auxll = auxll->pnext;
	auxll->pnext = freeLL;
	freeLL = first;
}
